import pandas as pd
import xarray as xr
import numpy as np
from scipy.integrate import cumulative_trapezoid as cumtrapz


from gfatpy.lidar.retrieval.klett import klett_rcs
from gfatpy.lidar.retrieval.raman import retrieve_backscatter, retrieve_extinction
from gfatpy.lidar.utils.utils import refill_overlap, signal_to_rcs
from gfatpy.atmo.rayleigh import molecular_properties


def overlap_iterative_raman(
    elastic_signal: xr.DataArray,
    raman_signal: xr.DataArray,
    meteo_profiles: pd.DataFrame,
    particle_lidar_ratio: float,
    wavelengths: tuple[float, float],
    reference: tuple[float, float],
    beta_aer_ref: float = 0,
    min_overlap_range: float = 200,  # minimum range of overlap function
    particle_angstrom_exponent: float = 0.0,
    iteration_limit: int = 10,
    debugging: bool = False,
    **kwargs: dict,
) -> xr.Dataset:
    """Retrieve overlap function by means of the Wandinger's method.

    Args:
        elastic_signal (xr.DataArray): Elastic lidar signal at wavelength `wavelengths[0]`.
        raman_signal (xr.DataArray): Raman lidar signal at wavelength `wavelengths[1]`.
        meteo_profiles (pd.DataFrame): from  `gfatpy.atmo.atmo.generate_meteo_profiles` with pressure and temperature data.
        particle_lidar_ratio (float): Particle lidar ratio to be used in the Klett inversion. It shall be chosen so Klett backscatter profile fits Raman backscatter profile once full overlap is reached.
        reference (tuple[float, float], optional): Reference range required by the Klett and Raman inversions. Defaults to (8000,8500).
        wavelengths (tuple[float, float], optional): Elastic and Raman wavelengths. Defaults to (532, 531).
        beta_aer_ref (float, optional): Particle backscatter coefficient at `reference range` in m^-1sr^-1. Defaults to 0 m^-1sr^-1.
        particle_angstrom_exponent (float, optional): Particle extinction-related Angstrom exponent. It is used in the Raman inversion. Defaults to 0.
        min_overlap_range (float, optional): Overlap function will be forced to zero below [m]. Defaults to 200 m.
        iteration_limit (int, optional): Maximum number of iterations. Defaults to 10.
        debugging (bool, optional): If True, it will return a dataset with all the intermediate variables. Defaults to False.
        **kwargs: Additional arguments such as `fulloverlap_height` and `output_dir`.

    Returns:
        xr.Dataset: overlap function. If debugging is True, it will return a dataset with all the intermediate variables.
    """
    # Check z in elastic and raman signals (the same for both signals)
    if len(elastic_signal.range) != len(raman_signal.range):
        raise ValueError(
            "elastic and Raman signals must have the same 'range' dimension."
        )

    if len(elastic_signal.range) != len(meteo_profiles["height"]):
        raise ValueError(
            "Signals range dimension and meteo_profiles['heights'] must have the same size."
        )

    z = elastic_signal.range.values.astype(np.float64)

    mol_properties = molecular_properties(
        wavelengths[0],
        meteo_profiles["pressure"],
        meteo_profiles["temperature"],
        heights=z,
    )
    molecular_backscatter = mol_properties["molecular_beta"].values.astype(np.float64)
    pressure = meteo_profiles["pressure"].values.astype(np.float64)
    temperature = meteo_profiles["temperature"].values.astype(np.float64)

    # Beta Klett
    elastic_backscater = klett_rcs(
        signal_to_rcs(elastic_signal, z).astype(np.float64),
        z,
        molecular_backscatter,
        lr_part=particle_lidar_ratio,
        reference=reference,
    )

    extinction = retrieve_extinction(
        raman_signal.values.astype(np.float64),
        z,
        wavelengths,
        pressure,
        temperature,
        reference=reference,
    )

    # Refill extinction
    extinction = refill_overlap(extinction, z, kwargs.get("fulloverlap_height", 600.0))

    raman_backscatter = retrieve_backscatter(
        raman_signal.values.astype(np.float64),
        elastic_signal.values.astype(np.float64),
        extinction,
        z,
        wavelengths,
        pressure,
        temperature,
        particle_angstrom_exponent=particle_angstrom_exponent,
    )  # Fine mode has been chosen in Amstrong exponent

    corrected_signal = np.zeros([iteration_limit, len(elastic_signal)])
    delta_overlap = np.zeros(corrected_signal.shape)
    elastic_backscatter_matrix = np.zeros(corrected_signal.shape)
    overlap_function = np.zeros(corrected_signal.shape)
    correction_factor = np.zeros(corrected_signal.shape)

    i = 0
    corrected_signal[i, :] = elastic_signal
    elastic_backscatter_matrix[i, :] = elastic_backscater
    correction_factor[i, :] = (
        2 * raman_backscatter + molecular_backscatter - elastic_backscatter_matrix[i, :]
    ) / (raman_backscatter + molecular_backscatter)

    # Iteraciones restantes
    for i in range(1, iteration_limit):
        corrected_signal[i, :] = (
            corrected_signal[i - 1, :] * correction_factor[i - 1, :]
        )
        rcs_ = signal_to_rcs(corrected_signal[i, :], z)
        elastic_backscatter_matrix[i, :] = klett_rcs(
            rcs_,
            z,
            molecular_backscatter,
            lr_part=particle_lidar_ratio,
            beta_aer_ref=beta_aer_ref,
            reference=reference,
        )

        correction_factor[i, :] = (
            2 * raman_backscatter
            + molecular_backscatter
            - elastic_backscatter_matrix[i, :]
        ) / (raman_backscatter + molecular_backscatter)

        overlap_function[i, :] = corrected_signal[0, :] / corrected_signal[i, :]

    final_overlap_function = overlap_function[-1, :]

    # Set to zero the values of overlap function below 200 m
    final_overlap_function[: np.where(z > min_overlap_range)[0][0]] = 0

    # Search first 1 in overlap function
    first_1 = np.where(final_overlap_function >= 1)[0][0]

    # Set constant profile to 1 form first 1 up
    final_overlap_function[first_1:] = 1

    final_overlap_function = xr.DataArray(
        final_overlap_function, dims=("range"), coords={"range": z}
    )

    if debugging:
        # Generate a xr.Dataset to save the results
        overlap_function = xr.DataArray(
            overlap_function,
            dims=("iteration", "range"),
            coords={"iteration": np.arange(iteration_limit), "range": z},
        )
        elastic_backscatter_matrix = xr.DataArray(
            elastic_backscatter_matrix,
            dims=("iteration", "range"),
            coords={"iteration": np.arange(iteration_limit), "range": z},
        )
        raman_backscatter = xr.DataArray(
            raman_backscatter, dims=("range"), coords={"range": z}
        )
        extinction = xr.DataArray(extinction, dims=("range"), coords={"range": z})
        correction_factor = xr.DataArray(
            correction_factor,
            dims=("iteration", "range"),
            coords={"iteration": np.arange(iteration_limit), "range": z},
        )
        delta_overlap = xr.DataArray(
            delta_overlap,
            dims=("iteration", "range"),
            coords={"iteration": np.arange(iteration_limit), "range": z},
        )
        corrected_signal = xr.DataArray(
            corrected_signal,
            dims=("iteration", "range"),
            coords={"iteration": np.arange(iteration_limit), "range": z},
        )
        dataset = xr.Dataset(
            {
                "overlap_function_matrix": overlap_function,
                "elastic_backscatter_matrix": elastic_backscatter_matrix,
                "raman_backscatter": raman_backscatter,
                "correction_factor": correction_factor,
                "delta_overlap": delta_overlap,
                "corrected_signal": corrected_signal,
                "extinction": extinction,
                "overlap_function": final_overlap_function,
            }
        )

    else:
        dataset = xr.Dataset(
            {
                "overlap_function": final_overlap_function,
            }
        )

    return dataset

def overlap_iterative_klett(
    ff_elastic_signal: xr.DataArray,
    nf_elastic_signal: xr.DataArray,
    meteo_profiles: pd.DataFrame,
    particle_lidar_ratio: float,
    wavelength: float,
    reference: tuple[float, float],
    beta_aer_ref: float = 0,
    min_overlap_range: float = 200,  # minimum range of overlap function
    iteration_limit: int = 10,
    debugging: bool = False,
    **kwargs: dict,
) -> xr.Dataset:
    """Retrieve overlap function by means of the iterative method with near-field Klett as reference.

    Args:
        ff_elastic_signal (xr.DataArray): Far-field elastic lidar signal.
        nf_elastic_signal (xr.DataArray): Near-field elastic lidar signal.
        meteo_profiles (pd.DataFrame): from  `gfatpy.atmo.atmo.generate_meteo_profiles` with pressure and temperature data.
        particle_lidar_ratio (float): Particle lidar ratio to be used in the Klett inversion. It shall be chosen so Klett backscatter profile fits Raman backscatter
        wavelength (float): Elastic wavelength.
        reference (tuple[float, float]): Reference range required by the Klett inversion.
        beta_aer_ref (float, optional): Particle backscatter coefficient at `reference range` in m^-1sr^-1. Defaults to 0 m^-1sr^-1.
        min_overlap_range (float, optional): Overlap function will be forced to zero below [m]. Defaults to 200 m.
        debugging (bool, optional): If True, it will return a dataset with all the intermediate variables. Defaults to False.

    Raises:
        ValueError: Both nf- and ff-elastic signals must have the same 'range' dimension.
        ValueError: Signals range dimension and meteo_profiles['heights'] must have the same size.

    Returns:
        xr.Dataset: overlap function. If debugging is True, it will return a dataset with all the intermediate variables.
    """    

    # Check z in elastic and raman signals (the same for both signals)
    if len(ff_elastic_signal.range) != len(nf_elastic_signal.range):
        raise ValueError(
            "Both nf- and ff-elastic signals must have the same 'range' dimension."
        )

    if len(ff_elastic_signal.range) != len(meteo_profiles["height"]):
        raise ValueError(
            "Signals range dimension and meteo_profiles['heights'] must have the same size."
        )

    z = ff_elastic_signal.range.values.astype(np.float64)

    mol_properties = molecular_properties(
        wavelength,
        meteo_profiles["pressure"],
        meteo_profiles["temperature"],
        heights=z,
    )
    molecular_backscatter = mol_properties["molecular_beta"].values.astype(np.float64)

    # Beta Klett
    ff_elastic_backscater = klett_rcs(
        signal_to_rcs(ff_elastic_signal, z).astype(np.float64),
        z,
        molecular_backscatter,
        lr_part=particle_lidar_ratio,
        reference=reference,
    )

    nf_elastic_backscater = klett_rcs(
        signal_to_rcs(nf_elastic_signal, z).astype(np.float64),
        z,
        molecular_backscatter,
        lr_part=particle_lidar_ratio,
        reference=reference,
    )

    corrected_signal = np.zeros([iteration_limit, len(ff_elastic_signal)])
    delta_overlap = np.zeros(corrected_signal.shape)
    ff_elastic_backscatter_matrix = np.zeros(corrected_signal.shape)
    overlap_function = np.zeros(corrected_signal.shape)
    correction_factor = np.zeros(corrected_signal.shape)

    i = 0
    corrected_signal[i, :] = ff_elastic_signal
    ff_elastic_backscatter_matrix[i, :] = ff_elastic_backscater
    correction_factor[i, :] = (
        2 * nf_elastic_backscater + molecular_backscatter - ff_elastic_backscatter_matrix[i, :]
    ) / (nf_elastic_backscater + molecular_backscatter)

    # Iteraciones restantes
    for i in range(1, iteration_limit):
        corrected_signal[i, :] = (
            corrected_signal[i - 1, :] * correction_factor[i - 1, :]
        )
        rcs_ = signal_to_rcs(corrected_signal[i, :], z)
        ff_elastic_backscatter_matrix[i, :] = klett_rcs(
            rcs_,
            z,
            molecular_backscatter,
            lr_part=particle_lidar_ratio,
            beta_aer_ref=beta_aer_ref,
            reference=reference,
        )

        correction_factor[i, :] = (
            2 * nf_elastic_backscater
            + molecular_backscatter
            - ff_elastic_backscatter_matrix[i, :]
        ) / (nf_elastic_backscater + molecular_backscatter)

        overlap_function[i, :] = corrected_signal[0, :] / corrected_signal[i, :]

    final_overlap_function = overlap_function[-1, :]

    # Set to zero the values of overlap function below 200 m
    final_overlap_function[: np.where(z > min_overlap_range)[0][0]] = 0

    # Search first 1 in overlap function
    first_1 = np.where(final_overlap_function >= 1)[0][0]

    # Set constant profile to 1 form first 1 up
    final_overlap_function[first_1:] = 1

    final_overlap_function = xr.DataArray(
        final_overlap_function, dims=("range"), coords={"range": z}
    )

    if debugging:
        # Generate a xr.Dataset to save the results
        overlap_function = xr.DataArray(
            overlap_function,
            dims=("iteration", "range"),
            coords={"iteration": np.arange(iteration_limit), "range": z},
        )
        ff_elastic_backscatter_matrix = xr.DataArray(
            ff_elastic_backscatter_matrix,
            dims=("iteration", "range"),
            coords={"iteration": np.arange(iteration_limit), "range": z},
        )
        nf_elastic_backscatter = xr.DataArray(
            nf_elastic_backscater, dims=("range"), coords={"range": z}
        )
        correction_factor = xr.DataArray(
            correction_factor,
            dims=("iteration", "range"),
            coords={"iteration": np.arange(iteration_limit), "range": z},
        )
        delta_overlap = xr.DataArray(
            delta_overlap,
            dims=("iteration", "range"),
            coords={"iteration": np.arange(iteration_limit), "range": z},
        )
        corrected_signal = xr.DataArray(
            corrected_signal,
            dims=("iteration", "range"),
            coords={"iteration": np.arange(iteration_limit), "range": z},
        )
        dataset = xr.Dataset(
            {
                "overlap_function_matrix": overlap_function,
                "elastic_backscatter_matrix": ff_elastic_backscatter_matrix,
                "nf_elastic_backscatter": nf_elastic_backscatter,
                "correction_factor": correction_factor,
                "delta_overlap": delta_overlap,
                "corrected_signal": corrected_signal,
                "overlap_function": final_overlap_function,
            }
        )

    else:
        dataset = xr.Dataset(
            {
                "overlap_function": final_overlap_function,
            }
        )

    return dataset


def overlap_function_explicit(
    elastic_rcs: xr.DataArray,
    raman_rcs: xr.DataArray,
    wavelengths: tuple[float, float],
    reference_heights: tuple[float, float],
    particle_lidar_ratio: float,
    meteo_profiles: pd.DataFrame,
    debugging=False,
    **kwargs: dict,
) -> np.ndarray:
    """Retrieve overlap function by means of the explicit method.

    Args:

        - elastic_rcs (xr.DataArray): Elastic range corrected signal at wavelength `wavelengths[0]`.
        - raman_rcs (xr.DataArray): Raman range corrected signal at wavelength `wavelengths[1]`.
        - wavelengths (tuple[float, float]): Elastic and Raman wavelengths.
        - reference_height (float): Aerosol-free region reference height in meters.
        - particle_lidar_ratio (float): Assumed particle lidar ratio.
        - meteo_profiles (pd.DataFrame): Meteo profiles with pressure and temperature data from `gfatpy.atmo.atmo.generate_meteo_profiles`.

    Returns:

        - xr.Dataset: overlap function. If debugging is True, it will return a dataset with all the intermediate variables.

    References:

        - Comerón et al., 2023: https://doi.org/10.5194/amt-16-3015-2023
    """

    r = elastic_rcs.range.values.astype(np.float64)
    dr = r[1] - r[0]

    # Create an array full of 1 of size r
    array_ones = np.ones(r.shape)
    elastic_mol_properties = molecular_properties(
        wavelengths[0],
        meteo_profiles["pressure"],
        meteo_profiles["temperature"],
        heights=r,
    )
    raman_mol_properties = molecular_properties(
        wavelengths[1],
        meteo_profiles["pressure"],
        meteo_profiles["temperature"],
        heights=r,
    )

    # Computation of beta molecular for elastic and Raman signal
    beta_mol_elastic = elastic_mol_properties["molecular_beta"]  # it has to be in m-1*sr-1
    beta_mol_raman = raman_mol_properties["molecular_beta"]

    # Cálculo inverso de beta molecular, a partir de la señal Rayleigh
    XE = elastic_rcs.values  # elastic signal
    X_Ray = XE  # molecular backscatter normalized to RCS

    XR = raman_rcs.values  # raman signal
    XR_Ray = XR  # molecular backscatter normalized to RCS

    ref_height = np.mean(np.array(reference_heights))  # Reference height (m)
    idx = np.round(ref_height / dr).astype(int)
    idx_min = np.round(reference_heights[0] / dr).astype(int)
    idx_max = np.round(reference_heights[1] / dr).astype(int)

    X_ref = np.nanmean(X_Ray[idx_min:idx_max])
    XR_ref = np.nanmean(XR_Ray[idx_min:idx_max])
    beta_mol_ref = np.nanmean(beta_mol_elastic[idx_min:idx_max])

    norm_XR = XR / XR_ref
    nbeta_molecular = beta_mol_elastic / beta_mol_ref

    Sa = particle_lidar_ratio  # Assumed Lidar Ratio
    Sm_el = elastic_mol_properties["molecular_lidar_ratio"]
    # Sm_ra = raman_mol_properties['molecular_lidar_ratio']

    beta_mol_el = beta_mol_elastic[:idx]
    beta_mol_ra = beta_mol_raman[:idx]
    nbeta_mol = nbeta_molecular[:idx]
    nXR = norm_XR[:idx]
    X = XE[:idx]

    fexp = np.exp(
        -2
        * (Sa - Sm_el.values)
        * np.flip(cumtrapz(np.flip(beta_mol_el), initial=0, dx=dr))
    )

    # Not used because its efect is negligible
    # fexp_dif_ext_mol = np.exp(- Sm_el * np.flip(cumtrapz(np.flip(beta_mol_el - beta_mol_ra), initial=0, dx=dr)))

    g = (nbeta_mol / nXR) * fexp

    phi = (2 * beta_mol_el / (nXR * X_ref)) * fexp

    psi = (X * Sa) / fexp

    phi_psi = phi * psi

    fexp_phi_psi = np.exp(np.flip(cumtrapz(np.flip(phi_psi), initial=0, dx=dr)))

    fint_g_psi = np.flip(cumtrapz(np.flip(g * psi / fexp_phi_psi), initial=0, dx=dr))

    f = g + phi * fexp_phi_psi * fint_g_psi

    Oe = 1 / f

    # Paste Oe into the array of ones so that it has the same length as r
    Oe = np.concatenate((Oe, array_ones[idx:]))

    # Create a dataset with all the intermediate parameters
    Oe = xr.DataArray(Oe, dims=("range"), coords={"range": r})

    if debugging:
        f_ = np.empty_like(r)
        f_[:idx] = f
        g_ = np.empty_like(r)
        g_[:idx] = g
        phi_ = np.empty_like(r)
        phi_[:idx] = phi
        psi_ = np.empty_like(r)
        psi_[:idx] = psi
        phi_psi_ = np.empty_like(r)
        fexp_ = np.empty_like(r)
        fexp_[:idx] = fexp
        phi_psi_[:idx] = phi_psi
        fexp_phi_psi_ = np.empty_like(r)
        fexp_phi_psi_[:idx] = fexp_phi_psi
        fint_g_psi_ = np.empty_like(r)
        fint_g_psi_[:idx] = fint_g_psi

        f = xr.DataArray(f_, dims=("range"), coords={"range": r})
        g = xr.DataArray(g_, dims=("range"), coords={"range": r})
        phi = xr.DataArray(phi_, dims=("range"), coords={"range": r})
        psi = xr.DataArray(psi_, dims=("range"), coords={"range": r})
        fexp = xr.DataArray(fexp_, dims=("range"), coords={"range": r})
        phi_psi = xr.DataArray(phi_psi_, dims=("range"), coords={"range": r})
        fexp_phi_psi = xr.DataArray(fexp_phi_psi_, dims=("range"), coords={"range": r})
        fint_g_psi = xr.DataArray(fint_g_psi_, dims=("range"), coords={"range": r})
        beta_mol_el = xr.DataArray(beta_mol_elastic, dims=("range"), coords={"range": r})
        beta_mol_ra = xr.DataArray(beta_mol_raman, dims=("range"), coords={"range": r})
        nbeta_mol = xr.DataArray(nbeta_molecular, dims=("range"), coords={"range": r})
        nXR = xr.DataArray(norm_XR, dims=("range"), coords={"range": r})
        X = xr.DataArray(XE, dims=("range"), coords={"range": r})
        XR = xr.DataArray(XR, dims=("range"), coords={"range": r})

        dataset = xr.Dataset(
            {
                "overlap_function": Oe,
                "f": f,
                "g": g,
                "phi": phi,
                "psi": psi,
                "phi_psi": phi_psi,
                "fexp_phi_psi": fexp_phi_psi,
                "fint_g_psi": fint_g_psi,
                "beta_mol_el": beta_mol_el,
                "beta_mol_ra": beta_mol_ra,
                "nbeta_mol": nbeta_mol,
                "nXR": nXR,
                "X": X,
                "fexp": fexp,
                "XR": XR,
                "XR_Ray": XR_Ray,
                "X_Ray": X_Ray,
            }
        )
    else:
        dataset = xr.Dataset({"overlap_function": Oe})

    return dataset


from pathlib import Path
from datetime import datetime as dt, timedelta
from pdb import set_trace
import re
import zipfile

from gfatpy.lidar.utils.file_manager import info2general_path, info2path
from gfatpy.lidar.utils.utils import filter_wildcard
from gfatpy.lidar.utils.utils import get_532_from_telescope, licel_to_datetime, to_licel_date_str

from gfatpy.utils.io import unzip_file
from gfatpy.lidar.utils.utils import LIDAR_INFO

from gfatpy.lidar.utils.types import LidarName, MeasurementType, Telescope

RAW_FIRST_LETTER = LIDAR_INFO["metadata"]["licel_file_wildcard"]


class Measurement:
    path: Path
    type: MeasurementType
    datetime: dt
    is_zip: bool = False
    unzipped_path: Path | None = None
    files: list[str] = []
    has_dc: bool = False
    dc_path: Path | None = None
    linked_measurements: list = []

    def __init__(self, path: Path, type: MeasurementType, **kwargs):
        self.path = path
        self.type = type
        self.is_zip = self.path.suffix.endswith("zip")
        self.datetime = self.extract_datetime()
        self.filenames = [file.name for file in self.get_filepaths()]
        self.sub_dirs = self.extract_folders()
        self.has_dc = self.check_for_dc()
        self.dc_path = self.get_dc()
        self.unzipped_path = None

    def extract_datetime(self) -> dt:
        """get datetime from the measurement path

        Returns:

            - dt: datetime object
        """        
        formatted_date = self.path.name.split(".")[0][3:]
        return dt.strptime(formatted_date, r"%Y%m%d_%H%M")

    # def extract_filenames(self, wildcard: str = r"\.\d+$") -> list:
    #     """Extract filenames from the measurement path

    #     Args:
    #         wildcard (str, optional): wildcard pattern to filter the files. Defaults to r"\.\d+$".

    #     Raises:
    #         Exception: No files found in the directory to meet the wildcard.

    #     Returns:
    #         list: list of filenames.
    #     """        
    #     # Extract files from ZIP or directory based on wildcard pattern
    #     files = []
    #     if self.is_zip:

    #         unzip_file( self.path, pattern_or_list=wildcard, destination=destination )

    #         with zipfile.ZipFile(self.path, "r") as zip_ref:
    #             file_list = zip_ref.namelist()
    #             files = [
    #                 Path(file).name
    #                 for file in file_list
    #                 if (not Path(file).is_dir() and re.search(wildcard, Path(file).name))
    #             ]
    #     elif self.path.is_dir():            
    #         files = [file.name for file in self.path.rglob(wildcard)]
    #         if len(files) == 0:
    #             raise Exception(f"No files found in {self.path} to meet the wildcard {wildcard}")
    #     return files

    def extract_folders(self) -> list[str]:
        """Extract sub-directories from the measurement path

        Returns:

            - list[str]: list of sub-directories.
        """        
        folders = []
        path = self.path  # Store path in a local variable for use in the method
        if self.is_zip:
            with zipfile.ZipFile(path, "r") as zip_ref:
                file_list = zip_ref.namelist()
                folders = [
                    file.split("/")[-2] for file in file_list if file.endswith("/")
                ]
        elif path.is_dir():
            folders = [f.name for f in path.iterdir() if f.is_dir()]
        return folders

    def get_dc(self) -> Path | None:
        """Get the path for the DC type measurement

        Returns:

            - Path | None: Path for the DC type measurement. None if it does not exist.
        """        
        # Get the path for the DC type measurement
        if self.type == MeasurementType.DC:
            return None
        dc_path = self.path.parent / self.path.name.replace(
            self.type, MeasurementType.DC
        )
        return dc_path if dc_path.exists() else None

    def check_for_dc(self) -> bool:
        """Check if the DC type measurement exists 

        Returns:

            - bool: True if the DC type measurement exists. False otherwise.
        """        
        if self.type == MeasurementType.DC:
            return False

        dc_path = self.path.parent / self.path.name.replace(
            self.type, MeasurementType.DC
        )
        return dc_path.exists()

    def extract_zip(
        self, pattern_or_list: str = r'\.\d+$', destination: Path | None = None
    ) -> None:
        """Extract the zip file

        Args:

            - pattern_or_list (str, optional): pattern or list of patterns. Defaults to r'\.\d+$'.
            - destination (Path | None, optional): Directory to extract files. Defaults to None (extract to the same directory as the zip file).
        """        
        self.unzipped_path = unzip_file( self.path, pattern_or_list=pattern_or_list, destination=destination )


    def has_target_date(self, date: dt) -> bool:
        """Check if the measurement has the target date

        Args:

            - date (dt): Target date

        Returns:

            - bool: True if the measurement has the target date. False otherwise.
        """        
        has_target_date = False
        for file in self.files:
            match = re.search(f"{RAW_FIRST_LETTER}{to_licel_date_str(date)}*", file)
            if match != None and match.span()[1] > 6: 
                has_target_date = True
                break
        return has_target_date

    def generate_nc_output_path(
        self,
        output_path: Path,
        lidar_name: str,
        telescope: Telescope,
        measurement_type: MeasurementType,
        signal_type: str,
    ) -> Path:
        """Generate the output path for the netCDF file

        Args:

            - output_path (Path): Directory to save the netCDF file.
            - lidar_name (str): Lidar name.
            - telescope (Telescope): Telescope object (see gfatpy.lidar.nc_convert.types.Telescope).
            - measurement_type (MeasurementType): Measurement type (eg., RS, DC, etc.).
            - signal_type (str): Signal type. 

        Returns:

            - Path: Output path for the netCDF file.
        """        
        return info2path(
            lidar_name=lidar_name,
            channel=get_532_from_telescope(telescope),
            measurement_type=measurement_type,
            signal_type=signal_type,
            date=self.datetime,
            dir=output_path,
        )

    def find_prev_paths(
        self, number_of_previous_days: int, lidar_name: LidarName, raw_dir: Path
    ) -> list[Path]:
        """Find the paths for the previous days

        Args:

             number_of_previous_days (int): Number of previous days to look for.
             lidar_name (LidarName): Lidar name (see gfatpy.lidar.nc_convert.types.LidarName).
             raw_dir (Path): Directory for the raw data.

        Returns:

             list[Path]: List of paths for the previous days.
        """        
        current_date = self.datetime
        prev_paths = [
            info2general_path(
                lidar_name.value,
                date=current_date - timedelta(days=n_day),
                data_dir=raw_dir,
            )
            for n_day in range(1, number_of_previous_days + 1)
        ]
        # Remove paths that do not exist
        previous_paths = [prev_path for prev_path in prev_paths if prev_path.exists()]
        return previous_paths

    def add_linked_measurements(self, measurements: list) -> None:
        """Add linked measurements to the measurement object.

        Args:

            - measurements (list): List of measurements.
        """        
        # Add linked measurements to the measurement object
        # Linked measurements are measurements with the same date and type

        combine_list = self.linked_measurements + measurements
        self.linked_measurements = list(set(combine_list))
        
    
    def get_filepaths(self, pattern_or_list: str = r"\.\d+$", within_period: tuple[dt, dt] | None = None ) -> set[Path]:
        """Get the files from a measurement

        Args:

            - measurement (Measurement): Measurement object

        Returns:

            - set[Path]: Set of licel files
        """

        # Get files from the measurement main path
        if self.is_zip:
            if self.unzipped_path is None:
                self.extract_zip(pattern_or_list=pattern_or_list)
            dir_ = self.unzipped_path
        else:
            dir_ = self.path

        if dir_ is None:
            raise Exception(f"No files found in {self.path} to meet the wildcard {pattern_or_list}")            

        found_files = set(filter_wildcard(dir_))

        # Get files from linked measurements
        for linked_measurement in self.linked_measurements:
            if linked_measurement.is_zip:
                if linked_measurement.unzipped_path is None:
                    linked_measurement.extract_zip(pattern_or_list=pattern_or_list)
                dir_ = linked_measurement.unzipped_path
            else:
                dir_ = linked_measurement.path
            found_files.update(filter_wildcard(dir_))

        # Filter files2convert by larger than date_ini and smaller than date_end
        if within_period is not None:
            date_ini, date_end = within_period
            selected_files = { licel_ for licel_ in found_files if date_ini <= licel_to_datetime(licel_.name) <= date_end }
        else:
            selected_files = found_files
        return selected_files

    def __str__(self):
        return f"Measurement Object\nPath: {self.path}\nType: {self.type}\nDatetime: {self.datetime}\nIs ZIP: {self.is_zip}\nFiles: {self.files}\nSub-directories: {self.sub_dirs}\nDC Path: {self.dc_path}\nHas DC: {self.has_dc}\nUnzipped Path: {self.unzipped_path}\n\nLinked Measurements: {self.linked_measurements}\n"



#!/usr/bin/env python
# coding: utf-8

# import csv

import os
import glob
import numpy as np
import pandas as pd
import xarray as xr
import datetime as dt
import netCDF4 as nc
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.dates import DateFormatter
import matplotlib.dates as mdates

__version__ = "1.0.0"
__author__ = "Juan Antonio Bravo-Aranda"

# script description
BASE_DIR = os.path.dirname(os.path.dirname(__file__))
PROG_NAME = "disdrometerConverter"
PROG_DESCR = "converting raw data to netCDF files."

###############################################################################################################
###############################################################################################################
###############################################################################################################


def mis2nc(MIS_FILE, NC_OUT_FILE):
    """
    Esta funcion convierte un archivo .mis preprocesado (limpio) a un archivo .nc
    Variables de entrada:
    1. MIS_FILE= Ruta del archivo de entrada .mis (str)
    2. NC_OUT_FILE= Ruta del archivo de salida .nc (str)
    """

    dat = pd.read_csv(MIS_FILE, header=None).values

    # CREA NC
    ncout = nc.Dataset(NC_OUT_FILE, "w", format="NETCDF4")

    # DEFINE DIMENSIONES
    time = ncout.createDimension(
        "time", None
    )  # filelen, set='none' if unlimited dimension
    dclasses = ncout.createDimension(
        "dclasses", 32
    )  # sorting into velocity classes = bins
    vclasses = ncout.createDimension(
        "vclasses", 32
    )  # sorting into diameter classes = bins

    # DEFINE ATRIBUTOS GLOBALES:
    ncout.Title = "Parsivel disdrometer data"
    ncout.Institution = "ANDALUSIAN INSTITUTE FOR EARTH SYSTEM RESEARCH (Granada,Spain)"
    ncout.Contact_person = "Dr. Juan Bravo (jabravo@ugr.es )"
    ncout.Source = "NETCDF4"
    ncout.History = "Data processed on python."
    ncout.Dependencies = "external"
    ncout.Conventions = "CF-1.6 where applicable"
    ncout.Processing_date = dt.datetime.today().strftime("%Y-%m-%d,%H:%m:%S")
    ncout.Author = "Irving Juanico (iejuv@ier.unam.mx)"
    ncout.Comments = ""
    ncout.Licence = "For non-commercial use only. These data are the property of IISTA, their use is strictly prohibited without the authorization of the institute"

    # DEFINE VARIABLES
    # Define dimensiones como variables
    time = ncout.createVariable("time", np.float64, ("time",))
    time.fill_value = np.nan
    time.units = "hours since 2000-01-01"
    time.long_name = "time"

    dclasses = ncout.createVariable("dclasses", np.float64, ("dclasses",))
    dclasses.fill_value = np.nan
    dclasses.units = "mm"
    dclasses.long_name = "volume equivalent diameter class center"

    vclasses = ncout.createVariable("vclasses", np.float64, ("vclasses",))
    vclasses.fill_value = np.nan
    vclasses.units = "m s-1"
    vclasses.long_name = "velocity class center"

    # Definir exclusivamente variables dependientes de tiempo, clases v y clases d
    N = ncout.createVariable(
        "droplet_number", np.float64, ("time", "dclasses", "vclasses")
    )
    N.fill_value = np.nan
    N.units = "-"
    N.long_name = "number of droplets per volume equivalent diameter class and velocity class center"

    # Definir variables exclusivamente dependientes del tiempo.

    #'RECORD'
    inpr = ncout.createVariable("record_number", np.float32, ("time",))
    inpr.fill_value = np.nan
    inpr.units = "#"
    inpr.long_name = "File record number."

    #'rainIntensity'
    inpr = ncout.createVariable("precipitation_rain", np.float32, ("time",))
    inpr.fill_value = np.nan
    inpr.units = "mm h-1"
    inpr.long_name = "Intensity rain precipitation"

    #'snowIntensity'
    snowprec = ncout.createVariable("precipitation_snow", np.float32, ("time",))
    snowprec.fill_value = np.nan
    snowprec.units = "mm h-1"
    snowprec.long_name = "Intensity snow precipitation"

    #'accPrec'
    pss = ncout.createVariable("precipitacion_accumulated", np.float32, ("time",))
    pss.fill_value = np.nan
    pss.units = "mm"
    pss.long_name = "Accumulated precipitation since start"

    #'weatherCodeWaWa'
    WaWa = ncout.createVariable("wmo_code_WaWa", np.float32, ("time",))
    WaWa.fill_value = np.nan
    WaWa.units = "-"
    WaWa.long_name = "weather code according to WMO SYNOP 4680"

    #'radarReflectivity'
    rr = ncout.createVariable("radar_reflectivity", np.float32, ("time",))
    rr.fill_value = np.nan
    rr.units = "dbZ"
    rr.long_name = "Radar reflectivity"

    #'morVisibility'
    MOR = ncout.createVariable("visibility_MOR", np.float32, ("time",))
    MOR.fill_value = np.nan
    MOR.units = "m"
    MOR.long_name = "MOR visibility in the precipitation"

    #'signalAmplitude'
    sal = ncout.createVariable("signal_amplitude", np.float32, ("time",))
    sal.fill_value = np.nan
    sal.units = "#"
    sal.long_name = "Signal amplitude of Laserband"

    #'numberParticles'
    Ndp = ncout.createVariable("detected_droplets", np.float32, ("time",))
    Ndp.fill_value = np.nan
    Ndp.units = "#"
    Ndp.long_name = "Number of detected particles"

    #'sensorTemperature'
    Ts = ncout.createVariable("sensor_temperature", np.float32, ("time",))
    Ts.fill_value = np.nan
    Ts.units = "ºC"
    Ts.long_name = "Temperature in sensor"

    #'heatingCurrent'
    Ih = ncout.createVariable("heating_current", np.float32, ("time",))
    Ih.fill_value = np.nan
    Ih.units = "A"
    Ih.long_name = "Heating current"

    #'sensorVoltage'
    Vs = ncout.createVariable("sensor_voltage", np.float32, ("time",))
    Vs.fill_value = np.nan
    Vs.units = "V"
    Vs.long_name = "Sensor Voltage"

    #'kineticEnergy'
    Ek = ncout.createVariable("kinetic_energy", np.float32, ("time",))
    Ek.fill_value = np.nan
    Ek.units = "J"
    Ek.long_name = "Kinetic energy"

    #'sensorStatus'
    Ss = ncout.createVariable("sensor_status", np.float32, ("time",))
    Ss.fill_value = np.nan
    Ss.units = "0: OK/ON  and   1: FUCK  2: OFF"
    Ss.long_name = "Sensor status"

    # EXTRAYENDO LOS DATOS (archivos .mis)

    date_list = dat[:, 0]
    date_nc = nc.date2num(
        [dt.datetime.strptime(date, "%Y-%m-%d %H:%M:%S") for date in date_list],
        time.units,
    )

    spectrum_list = dat[:, 14:-1]
    spectrum_matrix = np.array([np.split(spectrum, 32) for spectrum in spectrum_list])

    # ESCRIBIENDO NC

    dclasses[:] = np.array(
        [
            0.062,
            0.187,
            0.312,
            0.437,
            0.562,
            0.687,
            0.812,
            0.937,
            1.062,
            1.187,
            1.375,
            1.625,
            1.875,
            2.125,
            2.375,
            2.750,
            3.250,
            3.750,
            4.250,
            4.750,
            5.500,
            6.500,
            7.500,
            8.500,
            9.500,
            11,
            13,
            15,
            17,
            19,
            21.5,
            24.5,
        ]
    )
    vclasses[:] = np.array(
        [
            0.05,
            0.15,
            0.25,
            0.35,
            0.45,
            0.55,
            0.65,
            0.75,
            0.85,
            0.95,
            1.10,
            1.30,
            1.50,
            1.70,
            1.90,
            2.20,
            2.60,
            3.00,
            3.40,
            3.80,
            4.40,
            5.20,
            6.00,
            6.80,
            7.60,
            8.80,
            10.4,
            12.0,
            13.6,
            15.2,
            17.6,
            20.8,
        ]
    )
    time[:] = date_nc
    N[:] = spectrum_matrix
    inpr[:] = dat[:, 1]
    pss[:] = dat[:, 2]
    WaWa[:] = dat[:, 3]
    rr[:] = dat[:, 4]
    MOR[:] = dat[:, 5]
    sal[:] = dat[:, 6]
    Ndp[:] = dat[:, 7]
    Ts[:] = dat[:, 8]
    Ih[:] = dat[:, 9]
    Vs[:] = dat[:, 10]
    Ek[:] = dat[:, 11]
    Ss[:] = dat[:, 12]

    ncout.close()

    return


def dat2nc(DAT_FILE, NC_OUT_FILE):
    """
    Esta funcion convierte un archivo .dat a un archivo .nc
    Variables de entrada:
    1. DAT_FILE= Ruta del archivo de entrada .dat (str)
    2. NC_OUT_FILE= Ruta del archivo de salida .nc (str)
    """

    if os.path.isfile(DAT_FILE):
        df = pd.read_csv(
            DAT_FILE, skiprows=0, header=1, parse_dates=["TIMESTAMP"], low_memory=False
        )
        df = df.drop([0, 1])
        df["TIMESTAMP"] = pd.to_datetime(df["TIMESTAMP"])
        df.set_index("TIMESTAMP", inplace=True)

        # CREA NC
        ncout = nc.Dataset(NC_OUT_FILE, "w", format="NETCDF4")

        # DEFINE DIMENSIONES
        time = ncout.createDimension(
            "time", None
        )  # filelen, set='none' if unlimited dimension
        dclasses = ncout.createDimension(
            "dclasses", 32
        )  # sorting into velocity classes = bins
        vclasses = ncout.createDimension(
            "vclasses", 32
        )  # sorting into diameter classes = bins

        # DEFINE ATRIBUTOS GLOBALES:
        ncout.Title = "Parsivel disdrometer data"
        ncout.Institution = (
            "ANDALUSIAN INSTITUTE FOR EARTH SYSTEM RESEARCH (Granada,Spain)"
        )
        ncout.Contact_person = "Dr. Juan Bravo (jabravo@ugr.es )"
        ncout.Source = "NETCDF4"
        ncout.History = "Data processed on python."
        ncout.Dependencies = "external"
        ncout.Conventions = "CF-1.6 where applicable"
        ncout.Processing_date = dt.datetime.today().strftime("%Y-%m-%d,%H:%m:%S")
        ncout.Author = "Dr. Juan A. Bravo-Aranda (jabravo@ugr.es)"
        ncout.Comments = ""
        ncout.Licence = "For non-commercial use only. These data are the property of IISTA, their use is strictly prohibited without the authorization of the institute"

        # DEFINE VARIABLES
        # Define dimensiones como variables
        time = ncout.createVariable("time", np.float64, ("time",))
        time.fill_value = np.nan
        time.units = "hours since 2000-01-01 0:0:0"
        time.long_name = "time"

        dclasses = ncout.createVariable("dclasses", np.float64, ("dclasses",))
        dclasses.fill_value = np.nan
        dclasses.units = "mm"
        dclasses.long_name = "volume equivalent diameter class center"

        vclasses = ncout.createVariable("vclasses", np.float64, ("vclasses",))
        vclasses.fill_value = np.nan
        vclasses.units = "m s-1"
        vclasses.long_name = "velocity class center"

        # Definir exclusivamente variables dependientes de tiempo, clases v y clases d
        N = ncout.createVariable(
            "droplet_number", np.float64, ("time", "dclasses", "vclasses")
        )
        N.fill_value = np.nan
        N.units = "-"
        N.long_name = "number of droplets per volume equivalent diameter class and velocity class center"

        # Definir variables exclusivamente dependientes del tiempo.

        #'RECORD'
        inpr = ncout.createVariable("record_number", np.float32, ("time",))
        inpr.fill_value = np.nan
        inpr.units = "#"
        inpr.long_name = "File record number."

        #'rainIntensity'
        inpr = ncout.createVariable("precipitation_rain", np.float32, ("time",))
        inpr.fill_value = np.nan
        inpr.units = "mm h-1"
        inpr.long_name = "Intensity rain precipitation"

        #'snowIntensity'
        snowprec = ncout.createVariable("precipitation_snow", np.float32, ("time",))
        snowprec.fill_value = np.nan
        snowprec.units = "mm h-1"
        snowprec.long_name = "Intensity snow precipitation"

        #'accPrec'
        pss = ncout.createVariable("precipitacion_accumulated", np.float32, ("time",))
        pss.fill_value = np.nan
        pss.units = "mm"
        pss.long_name = "Accumulated precipitation since start"

        #'weatherCodeWaWa'
        WaWa = ncout.createVariable("wmo_code_WaWa", np.float32, ("time",))
        WaWa.fill_value = np.nan
        WaWa.units = "-"
        WaWa.long_name = "weather code according to WMO SYNOP 4680"

        #'radarReflectivity'
        rr = ncout.createVariable("radar_reflectivity", np.float32, ("time",))
        rr.fill_value = np.nan
        rr.units = "dbZ"
        rr.long_name = "Radar reflectivity"

        #'morVisibility'
        MOR = ncout.createVariable("visibility_MOR", np.float32, ("time",))
        MOR.fill_value = np.nan
        MOR.units = "m"
        MOR.long_name = "MOR visibility in the precipitation"

        #'signalAmplitude'
        sal = ncout.createVariable("signal_amplitude", np.float32, ("time",))
        sal.fill_value = np.nan
        sal.units = "#"
        sal.long_name = "Signal amplitude of Laserband"

        #'numberParticles'
        Ndp = ncout.createVariable("detected_droplets", np.float32, ("time",))
        Ndp.fill_value = np.nan
        Ndp.units = "#"
        Ndp.long_name = "Number of detected particles"

        #'sensorTemperature'
        Ts = ncout.createVariable("sensor_temperature", np.float32, ("time",))
        Ts.fill_value = np.nan
        Ts.units = "ºC"
        Ts.long_name = "Temperature in sensor"

        #'heatingCurrent'
        Ih = ncout.createVariable("heating_current", np.float32, ("time",))
        Ih.fill_value = np.nan
        Ih.units = "A"
        Ih.long_name = "Heating current"

        #'sensorVoltage'
        Vs = ncout.createVariable("sensor_voltage", np.float32, ("time",))
        Vs.fill_value = np.nan
        Vs.units = "V"
        Vs.long_name = "Sensor Voltage"

        #'kineticEnergy'
        Ek = ncout.createVariable("kinetic_energy", np.float32, ("time",))
        Ek.fill_value = np.nan
        Ek.units = "J"
        Ek.long_name = "Kinetic energy"

        #'sensorStatus'
        Ss = ncout.createVariable("sensor_status", np.float32, ("time",))
        Ss.fill_value = np.nan
        Ss.units = "0: OK/ON  and   1: FUCK  2: OFF"
        Ss.long_name = "Sensor status"

        #'rightTemperature'
        Tr = ncout.createVariable("temperature_right", np.float32, ("time",))
        Tr.fill_value = np.nan
        Tr.units = "ºC"
        Tr.long_name = "Temperature of the right arm."

        #'leftTemperature'
        Tl = ncout.createVariable("temperature_left", np.float32, ("time",))
        Tl.fill_value = np.nan
        Tl.units = "ºC"
        Tl.long_name = "Temperature of the left arm."

        #'pbcTemperature'
        Tpbc = ncout.createVariable("temperature_pbc", np.float32, ("time",))
        Tpbc.fill_value = np.nan
        Tpbc.units = "ºC"
        Tpbc.long_name = "pbc Temperature."

        #'errorCode'
        Ec = ncout.createVariable("error_code", np.float32, ("time",))
        Ec.fill_value = np.nan
        Ec.units = "ºC"
        Ec.long_name = "Error code."

        # EXTRAYENDO LOS DATOS (archivos .mis)
        date_nc = nc.date2num(df.index.to_pydatetime(), time.units)

        spectrum_array = df.iloc[:, -1024:].to_numpy()
        spectrum_matrix = np.array(
            [np.split(spectrum, 32) for spectrum in spectrum_array]
        )

        # ESCRIBIENDO NC
        time[:] = date_nc
        N[:] = spectrum_matrix
        inpr[:] = df["rainIntensity"].to_numpy()
        pss[:] = df["accPrec"].to_numpy()
        WaWa[:] = df["weatherCodeWaWa"].to_numpy()
        rr[:] = df["radarReflectivity"].to_numpy()
        MOR[:] = df["morVisibility"].to_numpy()
        sal[:] = df["signalAmplitude"].to_numpy()
        Ndp[:] = df["numberParticles"].to_numpy()
        Ts[:] = df["sensorTemperature"].to_numpy()
        Ih[:] = df["heatingCurrent"].to_numpy()
        Vs[:] = df["sensorVoltage"].to_numpy()
        Ek[:] = df["kineticEnergy"].to_numpy()
        Ss[:] = df["sensorStatus"].to_numpy()
        Tr[:] = df["rightTemperature"].to_numpy()
        Tl[:] = df["leftTemperature"].to_numpy()
        Tpbc[:] = df["pbcTemperature"].to_numpy()
        Ec[:] = df["errorCode"].to_numpy()

        dclasses[:] = np.array(
            [
                0.062,
                0.187,
                0.312,
                0.437,
                0.562,
                0.687,
                0.812,
                0.937,
                1.062,
                1.187,
                1.375,
                1.625,
                1.875,
                2.125,
                2.375,
                2.750,
                3.250,
                3.750,
                4.250,
                4.750,
                5.500,
                6.500,
                7.500,
                8.500,
                9.500,
                11,
                13,
                15,
                17,
                19,
                21.5,
                24.5,
            ]
        )
        vclasses[:] = np.array(
            [
                0.05,
                0.15,
                0.25,
                0.35,
                0.45,
                0.55,
                0.65,
                0.75,
                0.85,
                0.95,
                1.10,
                1.30,
                1.50,
                1.70,
                1.90,
                2.20,
                2.60,
                3.00,
                3.40,
                3.80,
                4.40,
                5.20,
                6.00,
                6.80,
                7.60,
                8.80,
                10.4,
                12.0,
                13.6,
                15.2,
                17.6,
                20.8,
            ]
        )

        ncout.close()
        print("Done!")
    return


def rawfile2nc(INPUT_PATH, OUTPUT_PATH):
    """
    rawfile2nc(INPUT_PATH,FILE_OUT_NC)
        This function converts a single *.mis or *.dat file in netCDF files.
    Input:
        1. INPUT_PATH= file path *.mis or *.dat (str)
        2. FILE_OUT_NC= (optional) output file path.(str)
    """
    print(INPUT_PATH)
    extension = INPUT_PATH.split(".")[-1]
    if extension == "mis":
        print("Converting mis-file: %s" % INPUT_PATH)
        mis2nc(INPUT_PATH, OUTPUT_PATH)
    elif extension == "dat":
        print("Converting dat-file: %s" % INPUT_PATH)
        dat2nc(INPUT_PATH, OUTPUT_PATH)
    else:
        print("File extension unknown: %s" % INPUT_PATH)
    return


def raw2nc(INPUT_PATH):
    """
    disdro2nc(INPUT_PATH)
        This function converts *.mis or *.dat files in netCDF files.
    Input:
        1. INPUT_PATH= file paths *.mis or *.dat (str)
    """

    file_list = glob.glob(INPUT_PATH)  # lista de archivos
    if len(file_list) >= 1:
        file_list.sort()  # Ordenada alfabéticamente los ficheros del directorio
        for file_ in file_list:
            print("Convirtiendo archivos...")
            if file_.find("1a") == -1:
                # parsivel_20200205.dat
                dtfile_ = dt.datetime.strptime(
                    file_.split("_")[1].split(".")[0], "%Y%m%d"
                )
                year = dtfile_.year
                month = dtfile_.month
                outputdir = os.path.join(
                    os.path.dirname(os.path.dirname(file_)),
                    "1a",
                    "%d" % year,
                    "%02d" % month,
                )
                filename = os.path.basename(file_).replace(".dat", ".nc")
                output_filepath = os.path.join(outputdir, filename)
            else:
                output_filepath = file_.replace("0a", "1b").replace(
                    file_.split(".")[-1], "nc"
                )
            if not os.path.exists(os.path.dirname(output_filepath)):
                os.makedirs(os.path.dirname(output_filepath))
            rawfile2nc(file_, output_filepath)
    return


###############################################################################################################


def spectrumPlot(
    station,
    spectrum_list,
    figurePath,
    daterange,
    plotrange=((0, 6.1), (0, 12)),
    size=17,
):
    """
    spectrumPlot(spectrum_list, daterange=tuple, plotrange=tuple, size):
        Esta función grafica la distribución de goteo del espectro medido.

    Variables de entrada son la siguientes:
    1. spectrum_list = Puede ser una lista de 1024 elementos o una matriz de [32,32] (np.array)
    2. daterange = Es una tupla definida como (date_0,date_f) donde date_o y date_son el rango
       inical y final de graficación respectivamente   (tuple of str)
    3. plotrange = Es una tupla definida como ((xmin,xmax),(ymin,ymax)) que permite modificar el
       rango de visualización en eje coordenado (x,y) del grafico, dicho rango se encuentra
       predefinido como plotrange=((0,6.1),(0,12))      (tuple of int/float)
    4. size = Tamaño de los encabezados (7 POR DEFECTO) (int/float)
    5. figurePath = ruta completa de la figura a guardar. (string)
        NOTA: las variables daterange[:] deben de ser escrita de la forma YYYY-MM-DD hh:mm:ss
           YYYY: Año escrito con 4 digitos
           MM: Numero del mes escrito con 2 digitos
           DD: Dia del mes escrito con 2 digitos
           hh: Hora del dia escrito con 2 digitos
           mm: Minuto de la hora escrito con 2 digitos
           ss: Segundos del minuto escrito con 2 digitos
    """
    if len(spectrum_list) == 1024:
        spectrum_list = np.split(spectrum_list, 32)
    else:
        spectrum_list = spectrum_list

    # Tamaño de la fuente de las etiquetas
    font = {"size": size}
    mpl.rc("font", **font)

    # Tamaño de la figura
    fig, axes = plt.subplots(nrows=1, figsize=(18, 10))

    # Colormap
    bounds = [
        0,
        1,
        5,
        25,
        50,
        100,
        250,
        500,
        1000,
        2000,
        4000,
        8000,
        16000,
        50000,
    ]  # Rangos del colormap discreto
    colors = [
        "#ffffff",
        "#0015ff",
        "#0051ff",
        "#0095ff",
        "#00ebfc",
        "#9ff8fc",
        "#e8fc60",
        "#f2ff00",
        "#fcac00",
        "#fc7e00",
        "#fc4700",
        "#fa0000",
        "#9c000f",
    ]  # colores de cada rango en formato HEX
    cm = mpl.colors.ListedColormap(colors)
    norm = mpl.colors.BoundaryNorm(bounds, cm.N)

    # Mallado del colormap
    dclasses = [
        0.062,
        0.187,
        0.312,
        0.437,
        0.562,
        0.687,
        0.812,
        0.937,
        1.062,
        1.187,
        1.375,
        1.625,
        1.875,
        2.125,
        2.375,
        2.750,
        3.250,
        3.750,
        4.250,
        4.750,
        5.500,
        6.500,
        7.500,
        8.500,
        9.500,
        11,
        13,
        15,
        17,
        19,
        21.5,
        24.5,
    ]  # Clases diametro
    vclasses = [
        0.05,
        0.15,
        0.25,
        0.35,
        0.45,
        0.55,
        0.65,
        0.75,
        0.85,
        0.95,
        1.10,
        1.30,
        1.50,
        1.70,
        1.90,
        2.20,
        2.60,
        3.00,
        3.40,
        3.80,
        4.40,
        5.20,
        6.00,
        6.80,
        7.60,
        8.80,
        10.4,
        12.0,
        13.6,
        15.2,
        17.6,
        20.8,
    ]  # Clases velocidad

    colormap = axes.pcolormesh(
        dclasses, vclasses, spectrum_list, cmap=cm, norm=norm, shading="gouraud"
    )  # genera el colormap

    # Etiquetas y rango de visualización
    axes.set_xlabel("Rain-droplet diameter, $m$$m$")
    axes.set_ylabel("Fall velocity, $m/s$")
    axes.set_xlim(plotrange[0][0], plotrange[0][1])
    axes.set_ylim(plotrange[1][0], plotrange[1][1])

    # Dibuja color map
    PLOT = plt.colorbar(colormap, ax=axes)
    PLOT.ax.set_ylabel("Rain-droplet number concentration, $\#/m^3$")

    # Dibuja el mallado del colormap
    xi, yi = np.meshgrid(dclasses, vclasses)
    axes.plot(xi, yi, "k-", alpha=0.3)  # Dibuja la lineas verticales del mallado
    axes.plot(xi.T, yi.T, "k-", alpha=0.3)  # Dibuja la lineas horizontales del mallado

    if daterange[0][:-9] == daterange[1][:-9]:
        date = daterange[0][:-9]
    else:
        date = daterange[0][:-9] + " to " + daterange[1][:-9]

    datetemp0 = dt.datetime.strptime(daterange[0], "%Y-%m-%d %H:%M:%S")
    datetemp1 = dt.datetime.strptime(daterange[1], "%Y-%m-%d %H:%M:%S")
    year0 = dt.datetime.strftime(datetemp0, "%Y")
    year1 = dt.datetime.strftime(datetemp1, "%Y")
    if year0 == year1:
        date0 = dt.datetime.strftime(datetemp0, "%H:%M %d/%m")
        date1 = dt.datetime.strftime(datetemp1, "%H:%M %d/%m/%y")
    else:
        date0 = dt.datetime.strftime(datetemp0, "%H:%M %d/%m/%y")
        date1 = dt.datetime.strftime(datetemp1, "%H:%M %d/%m/%y")

    axes.set_title(
        "Disdrometer DSD in %s | Period: %s - %s " % (station, date0, date1)
    )  #

    # Dibuja la linea teórica
    gunnKinzerX = [
        0.0783,
        0.0913,
        0.1064,
        0.1241,
        0.1447,
        0.1687,
        0.1966,
        0.229,
        0.267,
        0.312,
        0.363,
        0.424,
        0.494,
        0.576,
        0.671,
        0.783,
        0.913,
        1.064,
        1.241,
        1.447,
        1.687,
        1.966,
        2.29,
        2.67,
        3.12,
        3.63,
        4.24,
        4.94,
        5.76,
    ]
    gunnKinzerY = [
        0.18,
        0.25,
        0.32,
        0.4,
        0.47,
        0.57,
        0.7,
        0.87,
        1.03,
        1.21,
        1.46,
        1.7,
        2.03,
        2.36,
        2.74,
        3.22,
        3.72,
        4.24,
        4.76,
        5.24,
        5.90,
        6.42,
        7.08,
        7.65,
        8.18,
        8.62,
        8.93,
        9.08,
        9.17,
    ]
    plt.plot(gunnKinzerX, gunnKinzerY, "gray")

    fig.savefig(figurePath, bbox_inches="tight", dpi=100)
    plt.close(fig)
    if os.path.isfile(figurePath):
        print("Figure succesfully created: %s" % figurePath)
        control = True
    else:
        print("Figure NOT created")
        control = False
    return control


###############################################################################################################
###############################################################################################################
###############################################################################################################


def accumulatedSpectrum(
    mainpath,
    figuredir,
    daterange,
    station="UGR",
    delay=0,
    plotrange=((0, 6.1), (0, 12)),
    size=17,
):
    """
    accumulatedSpectrum(PATH, daterange=tuple, delay=2, plotrange=tuple, size=7):
    Esta función grafica la distribución de goteo acumulada en un intervalo de tiempo [dat_0,date_f]

    NOTA 1: El gráficador considera la horas de retraso de medición del disdrometro (2 horas)

    Variables de entrada:
        1. mainpath = Ruta principal de los fichero *.nc  (str)
        2. figuredir = Ruta principal de las figuras *.png  (str)
        3. daterange = Es una tupla definida como (date_0,date_f) donde date_o y date_son el rango
           inical y final de graficación respectivamente   (tuple of str, format '%Y-%m-%d %H:%M:%S')
        4. delay= Es el tiempo de retraso de medición en horas (2 HORAS POR DEFECTO) (int/float)
        5. plotrange = Es una tupla definida como ((xmin,xmax),(ymin,ymax)) que permite modificar el
           rango de visualización en eje coordenado (x,y) del grafico, dicho rango se encuentra
           predefinido como plotrange=((0,6.1),(0,12))      (tuple of int/float)
        6. size = Tamaño de los encabezados (7 POR DEFECTO) (int/float)
    NOTA 2: las variables daterange[:] deben de ser escrita de la forma YYYY-MM-DD hh:mm:ss
    YYYY: Año escrito con 4 digitos
    MM: Numero del mes escrito con 2 digitos
    DD: Dia del mes escrito con 2 digitos
    hh: Hora del dia escrito con 2 digitos
    mm: Minuto de la hora escrito con 2 digitos
    ss: Segundos del minuto escrito con 2 digitos
    """
    control = False
    TIME_DELAY = delay  # HOURS
    dateini = dt.datetime.strptime(daterange[0], "%Y-%m-%d %H:%M:%S") - dt.timedelta(
        hours=TIME_DELAY
    )
    dateend = dt.datetime.strptime(daterange[1], "%Y-%m-%d %H:%M:%S") - dt.timedelta(
        hours=TIME_DELAY
    )
    inidate = dt.datetime.strftime(dateini, "%Y%m%d-%H%M%S")
    enddate = dt.datetime.strftime(dateend, "%Y%m%d-%H%M%S")

    date_ = dateini
    PATH = []
    while date_ < dateend:
        year = dt.datetime.strftime(date_, "%Y")
        month = dt.datetime.strftime(date_, "%m")
        filename = "%s_%s.nc" % (dt.datetime.strftime(date_, "%Y%m%d"), station)
        testfile = os.path.join(mainpath, year, month, filename)
        if os.path.isfile(testfile):
            PATH.append(testfile)
            print("Existing file append!: %s" % testfile)
        else:
            filename = "parsivel_%s.nc" % (dt.datetime.strftime(date_, "%Y%m%d"))
            testfile = os.path.join(mainpath, year, month, filename)
            if os.path.isfile(testfile):
                PATH.append(testfile)
                print("Existing file append!: %s" % testfile)

        date_ = date_ + dt.timedelta(days=1)

    if PATH:
        # Read files
        data_nc = xr.open_mfdataset(PATH)

        # Select region to plot
        data_dates = data_nc.sel(time=slice(*[dateini, dateend]))

        spectrum_accumulate_matrix = np.nansum(data_dates.droplet_number.values, axis=0)

        # Figure output path
        figurename = "accuDSD_%s_%s_%s.png" % (station, inidate, enddate)
        figurepath = os.path.join(figuredir, figurename)

        # Check folder exists
        if not os.path.isdir(os.path.dirname(figurepath)):
            os.mkdir(os.path.dirname(figurepath))

        # Figure plot
        control = spectrumPlot(
            station,
            spectrum_accumulate_matrix,
            figurepath,
            daterange=daterange,
            plotrange=plotrange,
            size=size,
        )

    return control


###############################################################################################################
###############################################################################################################
###############################################################################################################


def quicklook(
    variables2plot,
    mainpath,
    figuredir,
    daterange,
    station="UGR",
    delay=0,
    plotrange=((0, 6), (0, 12)),
    axesTime="Default",
    size=16,
):
    """
     quicklook(variables2plot, mainpath, figuredir, station, daterange, delay=0, plotrange=((0,6),(0,12)),axesTime='Default',size=16):

    Esta función realiza las siguientes graficas del disdrometro:
        1. Distribución de goteo por diametro
        2. Distribución de goteo por velocidad

    NOTA 1: El gráficador puede considerar las horas de retraso de medición del disdrometro (2 horas)

    Variables de entrada:
        1. variables2plot = Variable to make the quicklook. Options: 'diameter' and 'velocity' (str)
        2. PATH= Ruta principal del fichero netcdf (str)
        2. daterange = Es una tupla definida como (date_0,date_f) donde date_o y date_son el rango
           inical y final de graficación respectivamente   (tuple of str)
        3. plotrange = Es una tupla definida como ((dmin,dmax),(vmin,vmax)) que permite modificar el
           rango de visualización de cada distribución, (dmin,dmax) para modificar el rango de la
           distribución de goteo por diametro y (vmin,vmax) para modificar el rango de la distribución
           de goteo por velocidad.
           predefinido como plotrange=((0,6),(0,12))      (tuple of int/float)
        4. delay= Es el tiempo de retraso de medición en horas (0 HORAS POR DEFECTO) (int/float)
        5. size = Tamaño de los encabezados (16 POR DEFECTO) (int/float)

    NOTA 2: las variables date_0 y date_f deben de ser escrita de la forma YYYY-MM-DD hh:mm:ss
           YYYY: Año escrito con 4 digitos
           MM: Numero del mes escrito con 2 digitos
           DD: Dia del mes escrito con 2 digitos
           hh: Hora del dia escrito con 2 digitos
           mm: Minuto de la hora escrito con 2 digitos
           ss: Segundos del minuto escrito con 2 digitos
    """

    def quicklook_DSD(
        var_, tclasses, data_dates, daterange, plotrange, station, figurepath
    ):
        # FIGURA 1: DISTRIBUCIÓN DE GOTEO POR DIAMETRO
        # Tamaño de la figura
        control = False
        dclasses = [
            0.062,
            0.187,
            0.312,
            0.437,
            0.562,
            0.687,
            0.812,
            0.937,
            1.062,
            1.187,
            1.375,
            1.625,
            1.875,
            2.125,
            2.375,
            2.750,
            3.250,
            3.750,
            4.250,
            4.750,
            5.500,
            6.500,
            7.500,
            8.500,
            9.500,
            11,
            13,
            15,
            17,
            19,
            21.5,
            24.5,
        ]  # Clases diametro
        vclasses = [
            0.05,
            0.15,
            0.25,
            0.35,
            0.45,
            0.55,
            0.65,
            0.75,
            0.85,
            0.95,
            1.10,
            1.30,
            1.50,
            1.70,
            1.90,
            2.20,
            2.60,
            3.00,
            3.40,
            3.80,
            4.40,
            5.20,
            6.00,
            6.80,
            7.60,
            8.80,
            10.4,
            12.0,
            13.6,
            15.2,
            17.6,
            20.8,
        ]  # Clases velocidad
        var2plot = {"diameter": dclasses, "velocity": vclasses}
        ylabelstr = {
            "diameter": "Rain-droplet diameter, $m$$m$",
            "velocity": "Rain-droplet fall velocity, $m/s$",
        }
        ylimDict = {"diameter": plotrange[0], "velocity": plotrange[1]}
        # Extrayendo matriz del espectro por todo el lapso de tiempo
        spectrum_matrix = data_dates.droplet_number.values
        # Caculando distribución de goteo por diametro y por velocidad
        spectrum_per_diameter = np.nansum(spectrum_matrix, 1)
        spectrum_per_velocity = np.nansum(spectrum_matrix, 2)
        I = spectrum_per_diameter.shape[1]
        J = spectrum_per_velocity.shape[1]

        # Artilugio para poder graficar las distribuciones
        droplet_per_diameter = np.array(
            [list(spectrum_per_diameter[:, i]) for i in range(I)]
        )
        droplet_per_velocity = np.array(
            [list(spectrum_per_velocity[:, i]) for i in range(J)]
        )

        DSD = {"diameter": droplet_per_diameter, "velocity": droplet_per_velocity}

        # Tamaño de la fuente de las etiquetas
        font = {"size": size}
        mpl.rc("font", **font)

        fig, axes = plt.subplots(nrows=1, figsize=(18, 10))

        # Colormap
        bounds = [
            0,
            1,
            2,
            5,
            10,
            20,
            50,
            100,
            200,
            500,
            700,
            1000,
            1200,
            1500,
        ]  # Rangos del colormap discreto
        colors = [
            "#ffffff",
            "#0015ff",
            "#0051ff",
            "#0095ff",
            "#00ebfc",
            "#9ff8fc",
            "#e8fc60",
            "#f2ff00",
            "#fcac00",
            "#fc7e00",
            "#fc4700",
            "#fa0000",
            "#9c000f",
        ]  # colores de cada rango en formato HEX
        cm = mpl.colors.ListedColormap(colors)
        norm = mpl.colors.BoundaryNorm(bounds, cm.N)
        colormap = axes.pcolormesh(
            tclasses, var2plot[var_], DSD[var_], cmap=cm, norm=norm, shading="gouraud"
        )  # genera el colormap

        # Dibuja color map
        PLOT = plt.colorbar(colormap, ax=axes)
        PLOT.ax.set_ylabel("Rain-droplet number concentration, $m^3$")

        # Dibuja el mallado del colormap
        #         xi, yi = np.meshgrid(tclasses, dclasses)
        #         axes.plot(xi.T, yi.T, 'k-', alpha=0.3) # Dibuja la lineas horizontales del mallado
        # Etiquetas
        plt.grid(which="both")
        axes.set_ylabel(ylabelstr[var_])

        if axesTime == "Default":
            axes.set_xlabel("Time, $HH:MM$")
            # intervalo tiempo (step) en eje del tiempo (step de 1 hora)
            hours = mdates.HourLocator(interval=1)
            axes.xaxis.set_major_locator(hours)
            # unidades del eje de tiempo
            h_fmt = mdates.DateFormatter("%H:%M")
            axes.xaxis.set_major_formatter(h_fmt)
        elif axesTime == "Automatic":
            axes.set_xlabel("Time")

        # rango de visualiación
        datetemp0 = dt.datetime.strptime(daterange[0], "%Y-%m-%d %H:%M:%S")
        datetemp1 = dt.datetime.strptime(daterange[1], "%Y-%m-%d %H:%M:%S")

        axes.set_xlim(datetemp0, datetemp1)
        axes.set_ylim(ylimDict[var_][0], ylimDict[var_][1])

        diff = datetemp1 - datetemp0
        if diff <= dt.timedelta(hours=24):
            hour_resolution = 1
            hour_fmt = "%H:%M"
        elif np.logical_and(
            diff > dt.timedelta(hours=24), diff <= dt.timedelta(hours=48)
        ):
            hour_resolution = 3
            hour_fmt = "%H:%M"
        elif np.logical_and(
            diff > dt.timedelta(hours=48), diff <= dt.timedelta(hours=72)
        ):
            hour_resolution = 6
            hour_fmt = "%d %H:%M"
        else:
            hour_resolution = 12
            hour_fmt = "%d/%m %Hh"
        xdates = np.arange(datetemp0, datetemp1, dt.timedelta(hours=hour_resolution))
        plt.xticks(xdates)
        axes.xaxis.set_major_formatter(DateFormatter(hour_fmt))
        fig.autofmt_xdate()

        year0 = dt.datetime.strftime(datetemp0, "%Y")
        year1 = dt.datetime.strftime(datetemp1, "%Y")
        if year0 == year1:
            date0 = dt.datetime.strftime(datetemp0, "%H:%M %d/%m")
            date1 = dt.datetime.strftime(datetemp1, "%H:%M %d/%m/%y")
        else:
            date0 = dt.datetime.strftime(datetemp0, "%H:%M %d/%m/%y")
            date1 = dt.datetime.strftime(datetemp1, "%H:%M %d/%m/%y")
        axes.set_title(
            "%s-DSD quicklook | %s Period: %s - %s " % (var_, station, date0, date1)
        )  #

        fig.savefig(figurepath, bbox_inches="tight", dpi=100)
        if os.path.isfile(figurepath):
            print("Figure succesfully created: %s" % figurepath)
            control = True
        else:
            print("Figure NOT created")
            control = False
        plt.close(fig)

        return control

    # Main code
    control = False
    TIME_DELAY = delay  # HOURS
    dateini = dt.datetime.strptime(daterange[0], "%Y-%m-%d %H:%M:%S") - dt.timedelta(
        hours=TIME_DELAY
    )
    dateend = dt.datetime.strptime(daterange[1], "%Y-%m-%d %H:%M:%S") - dt.timedelta(
        hours=TIME_DELAY
    )
    inidate = dt.datetime.strftime(dateini, "%Y%m%d-%H%M%S")
    enddate = dt.datetime.strftime(dateend, "%Y%m%d-%H%M%S")

    date_ = dateini
    PATH = []
    while date_ < dateend:
        year = dt.datetime.strftime(date_, "%Y")
        month = dt.datetime.strftime(date_, "%m")
        filename = "%s_%s.nc" % (dt.datetime.strftime(date_, "%Y%m%d"), station)
        testfile = os.path.join(mainpath, year, month, filename)
        if os.path.isfile(testfile):
            PATH.append(testfile)
            print("Existing file append!: %s" % testfile)
        else:
            filename = "parsivel_%s.nc" % (dt.datetime.strftime(date_, "%Y%m%d"))
            testfile = os.path.join(mainpath, year, month, filename)
            if os.path.isfile(testfile):
                PATH.append(testfile)
                print("Existing file append!: %s" % testfile)

        date_ = date_ + dt.timedelta(days=1)

    if PATH:
        # Read files
        data_nc = xr.open_mfdataset(PATH)

        # Select region to plot
        data_dates = data_nc.sel(time=slice(*[dateini, dateend]))

        # Mallado del colormap
        tclasses = (
            data_dates.time.values + TIME_DELAY * 60 * 60000000000
        )  # Clases tiempo

        for var_ in variables2plot:
            # Figure output path
            figurename = "quicklook_%s_%s_%s_%s.png" % (var_, station, inidate, enddate)
            figurepath = os.path.join(figuredir, figurename)

            # Check folder exists
            if not os.path.isdir(os.path.dirname(figurepath)):
                os.makedirs(os.path.dirname(figurepath))

            print("Plotting %s" % var_)
            control = quicklook_DSD(
                var_, tclasses, data_dates, daterange, plotrange, station, figurepath
            )
            if not control:
                print("ERROR: figure not created: %s" % figurepath)
    return control


###############################################################################################################
###############################################################################################################
###############################################################################################################


def main():
    # parser = OptionParser(usage="usage %prog [options]",
    #     version="%prog 1.0")
    # parser.add_option("-s", "--station_name",
    #     action="store",
    #     dest="station",
    #     default="UGR",
    #     help="Measurement station [default: %default].")
    # parser.add_option("-d", "--date2plot",
    #     action="store",
    #     dest="date2plot",
    #     default="20191121",
    #     help="Date to plot [default: %default].")
    # parser.add_option("-o", "--dir_out",
    #     action="store",
    #     dest="dir_out",
    #     default=".",
    #     help="Output folder [default: %default].")
    # parser.add_option("-i", "--dir_in",
    #     action="store",
    #     dest="dir_in",
    #     default=".",
    #     help="Input folder [default: %default].")
    # (options, args) = parser.parse_args()
    # dir0 = options.dir_in
    # dir2 = options.dir_out
    # strdate = options.date2plot
    # station = options.station
    # date_ = dt.datetime.strptime(strdate,'%Y%m%d')
    # year = dt.datetime.strftime(date_,'%Y')
    # month = dt.datetime.strftime(date_,'%m')
    # file0a = os.path.join(dir0, year, month, '%s_%s.mis' % (dt.datetime.strftime(date_,'%Y%m%d'), station))
    # if os.path.isfile(file0a):
    #     print('%s found!' % file0a)
    #     file1a = file0a.replace('0a', '1a')
    #     disdroconverter1b.cleanMIS(file0a, file1a)

    #     if os.path.isfile(file1a):
    #         print('%s found!' % file1a)
    #         file1b = file1a.replace('1a', '1b')
    #         file1b = file1b.replace('mis', 'nc')
    #         to_nc(file1a, file1b)

    #         if os.path.isfile(file1b):
    #             print('%s successfully converted!' % file1b)
    #         else:
    #             print('%s conversion to 1b-level FAILED!' % file0a)
    #     else:
    #         print('%s conversion to 1a-level FAILED!' % file0a)
    # else:
    #     print('%s not found.' % file0a)
    return


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding: utf-8



import csv
import glob
import numpy as np
import os
import pandas as pd
import datetime as dt


TO = '00:00:00'
TF = '23:59:00'


#---------------------------------------------------------------------------------------#
# Esta función elimina listas vacías de un archivo .mis
def deleteSpaceWhite(file):
    archive = open(file)                     # file es la ruta del archivo .mis
    archive_list = list(csv.reader(archive)) # Escribe el archivo .mis
    DEL = []
    for i in archive_list:
        if i!=[]:                            # Elimina listas vacías en el archivo
            DEL.append(i)
    return DEL                               # Regresa el archivo .mis en un lista sin LISTAS VACÌAS


#---------------------------------------------------------------------------------------#
# Esta función detecta si  existe un error de medición en archivo .mis, principalmente errores de falta y/o 
#excedente de elementos de la medición (si sobra o falta algùn dato), imprime las lineas de erros del archivo.
def checkmis(list_mis):
    lens = list(map(len,list_mis))           # longitud de los datos de cada medición
    check = []                               # lista vacía
    linea = 0                                # contador para saber línea de error

    for i in lens:
        linea = linea + 1 
        if i==15 or i==1039:                 # i=15 significa una medición sin lluvia, i=1039 una medición con 
                                             # lluvia
            check.append(1)
        else:
            check.append(0)
            print('ERROR en linea',linea,'  <--------------------------------------------------------- E R R O R')
                                                          # Imprime error y el numero de línea del error
    if sum(check)/len(check)==1:
        print('Los archivos estan completos \n ...')
    else:
        print('Los archivos NO estan completos \n ...')
    

#---------------------------------------------------------------------------------------#
# Esta función detecta si existe algun archivo .mis (pertenciente a un directorio) con errores de escritura e
# imprime los erros de linea de cada archivo .mis (si los hay).

def checkParsivel_directory(file):           # file es el directorio de la carpeta de archivos .mis
    print('PATH =', file, '\n')
    print('Buscando errores... \n ...')
    entries = os.listdir(file)               # lista de archivos .mis contenidos en el directorio(carpeta)
    entries.sort()
    files_mis = [file+i for i in entries]    # lista de directorios de cada archivo .mis
    for i in files_mis:
        print(i)
        mis = deleteSpaceWhite(i)            # Aplica función deleteSpaceWhite()
        checkmis(mis)                        # Aplica función checkmis()
    print('No hay ningun error!\n')
    return 


#---------------------------------------------------------------------------------------#
# Esta función convierte la fecha de medición al formato YYYY-MM-DD hh:mm:ss
def joinDate(date_list):
    YYMMDD = date_list[0][6:] + date_list[0][2] + date_list[0][3:6] + date_list[0][0:2] 
    hhmmss = date_list[1]
    JOIN = [YYMMDD.replace('.','-') + ' ' + hhmmss]
    return JOIN


#---------------------------------------------------------------------------------------#
def cleanVariables(list_variables):
# Esta función cambia el tipo de dato de la varibles medidas, las cuáles se encuentran en un tipo de dato 'str' 
# y lo reescribe a valores de tipo 'float'
    VAR = [float(i) for i  in list_variables]
    return VAR


#---------------------------------------------------------------------------------------#
# Esta función cambia el espectro medido, el cuál se encuentra en un tipo de dato 'str' y lo reescribe 
# a valores de tipo 'float'
def cleanSpectrum(spectrum_list):
    SPT = []
    for i in spectrum_list:
        if i=='':
            SPT.append(0.)
        else:
            SPT.append(float(i))
    return SPT


#---------------------------------------------------------------------------------------#
# Esta funcòn aplana una lista de listas una de una dimensión a un lista de una dimensión
def flatten(list_of_list):
    FLAT = []
    for i in list_of_list:
        for j in i:
            FLAT.append(j)
    return FLAT

################################################################################################################
################################################################################################################
################################################################################################################
DATA = flatten([[0]*11,[2],['<SPECTRUM>'],[0]*1024,['<SPECTRUM>']])

def generateDatelist(dates,TYPE='datetime.datetime',step=1):
    DATES = []
    if type(dates[0]) == type(dates[1]) == dt.datetime:
        date_0 = dates[0]
        date_f = dates[1]
    elif type(dates[0]) == type(dates[1]) == str:
        date_0 = dt.datetime.strptime(dates[0],'%Y-%m-%d %H:%M:%S')
        date_f = dt.datetime.strptime(dates[1],'%Y-%m-%d %H:%M:%S')
    else:
        print('Error: Input the elements of the tuple "dates" as str type or as datetime.datetime type.')

    date_difference = date_f -date_0
    date_difference = int(date_difference.seconds/60)

    DATES.append(date_0)
    while date_0 < date_f:
        date_i = date_0 + dt.timedelta(minutes=step)
        DATES.append(date_i)
        date_0 = date_i

    if TYPE == 'datetime.datetime':
        OUTPUT = DATES
    elif TYPE == 'str':
        OUTPUT = [ dt.datetime.strftime(i,'%Y-%m-%d %H:%M:%S') for i in DATES ]
              
    return OUTPUT

################################################################################################################
################################################################################################################
################################################################################################################

def complete_between_T0andTF(mis,date_time):
    mis = [ list(i) for i in mis]
    
    CLEAN = []
    CLEAN.append(mis[0])
    for i in range(len(mis)-1):
        date_difference = date_time[i+1] - date_time[i]
        DT = date_difference.seconds/60

        if DT == 1:
                CLEAN.append(mis[i+1])
        elif DT > 1:
            date_unknown = generateDatelist((date_time[i],date_time[i+1]),TYPE='str')[1:-1]
            for j in date_unknown:
                add = [j] + DATA
                CLEAN.append(add)

            CLEAN.append(mis[i+1])
            
    mis = np.array(CLEAN)
        
    return mis

################################################################################################################
################################################################################################################
################################################################################################################

def complete_noTF(mis,date_mis):
    
    CLEAN = []
    
    date_mis.append(date_mis[0][:10] + ' ' + TF)
    date_time = [ dt.datetime.strptime(i,'%Y-%m-%d %H:%M:%S') for i in date_mis]

    date_difference = date_time[-1] - date_time[-2]
    DT = date_difference.seconds/60
    date_unknown = generateDatelist((date_time[-2],date_time[-1]),TYPE='str')[1:]
    for i in date_unknown:
        add = [i] + DATA
        CLEAN.append(add)
    CLEAN = [list(i) for i in mis] + CLEAN

    mis = pd.DataFrame(CLEAN).values
    return mis


################################################################################################################
################################################################################################################
################################################################################################################

def complete_noTO(mis,date_mis):
    
    CLEAN = []
    
    date_mis = [date_mis[0][:10] + ' ' + TO] + date_mis
    date_time = [ dt.datetime.strptime(i,'%Y-%m-%d %H:%M:%S') for i in date_mis]
    
    date_difference = date_time[1] - date_time[0]
    DT = date_difference.seconds/60
    date_unknown = generateDatelist((date_time[0],date_time[1]),TYPE='str')[:-1]
    for i in date_unknown:
        add = [i] + DATA
        CLEAN.append(add)
    CLEAN = CLEAN + [list(i) for i in mis]
    
    mis = pd.DataFrame(CLEAN).values
    return mis

################################################################################################################
################################################################################################################
################################################################################################################

def complete_noTOandTF(mis,date_mis):
    
    mis = complete_noTO(mis,date_mis)
    date_mis = list(mis[:,0])
    
    mis = complete_noTF(mis,date_mis)
    
    return mis

################################################################################################################
################################################################################################################
################################################################################################################

def complete_data(mis,TO='00:00:00',TF='23:59:00',DATA=flatten([[0]*11,[2],['<SPECTRUM>'],[0]*1024,['<SPECTRUM>']])):

    date_mis = list(mis[:,0])

    #STRING = ['<SPECTRUM>']
    #TO = '00:00:00'
    #TF = '23:59:00'
    #print(len(mis),'\n',mis.shape) 
    #DATA = flatten([[0]*11,[2],STRING,[0]*1024,STRING])

    if date_mis[0][11:] == TO and date_mis[-1][11:] == TF:
        date_time = [ dt.datetime.strptime(i,'%Y-%m-%d %H:%M:%S') for i in date_mis]
        #print(1)
    elif date_mis[0][11:] != TO and date_mis[-1][11:] == TF:

        mis = complete_noTO(mis,date_mis)
        date_mis = list(mis[:,0])
        date_time = [ dt.datetime.strptime(i,'%Y-%m-%d %H:%M:%S') for i in date_mis]

        #print(2)
    elif date_mis[0][11:] == TO and date_mis[-1][11:] != TF:

        mis = complete_noTF(mis,date_mis)
        date_mis = list(mis[:,0])
        date_time = [ dt.datetime.strptime(i,'%Y-%m-%d %H:%M:%S') for i in date_mis]

        #print(3)
    elif date_mis[0][11:] != TO and date_mis[-1][11:] != TF:

        mis = complete_noTOandTF(mis,date_mis)
        date_mis = list(mis[:,0])
        date_time = [ dt.datetime.strptime(i,'%Y-%m-%d %H:%M:%S') for i in date_mis]

        #print(4)

    mis = complete_between_T0andTF(mis,date_time)

    #print(len(mis),'\n',mis.shape) 
    
    return mis

################################################################################################################
################################################################################################################
################################################################################################################

#---------------------------------------------------------------------------------------#
# Esta función limpia un archivo .mis dado su fichero y lo envia (almacena) en FILE_OUT 
def cleanMIS(file,file_out):
    #FILE_OUT = file.replace('0a','1a') # <---------------
                                       # Para cambiar la dirección donde se va guardar el archivo, escriba aqui la 
                                       # ruta, de lo contrario se guardara a la carpeta /1a
        
    mis = deleteSpaceWhite(file) # llamando función 'deleteSpaceWhite()' 
    MIS = []
    for i in mis:
        dates = joinDate(i[:2]) # llamando función 'joinDate()'
        variables = cleanVariables(i[2:14]) # llamando función 'cleanVariables()' 
        string0 = ['<SPECTRUM>']
        if i[-1]=='<SPECTRUM>ZERO</SPECTRUM>':
            spectrum = [0]*1024 # lista de ceros con longitud de 1024 elementos

        else:
            spectrum = cleanSpectrum(i[15:-1]) # llamando función 'cleanSpectrum()'
            spectrum.append(0.)

        measurement = flatten([dates,variables,string0,spectrum,string0]) # llamando función 'flatten()'
        MIS.append(measurement)
    
    if len(MIS) == 1440:
    	df_mis = pd.DataFrame(MIS)
    	df_mis.to_csv(file_out,header=None,index=None)
    elif len(MIS) != 1440:
    	mis = pd.DataFrame(MIS).values
    	MIS = complete_data(mis)
    	df_mis = pd.DataFrame(MIS)
    	df_mis.to_csv(file_out,header=None,index=None) # exporta archivo a .mis

    return 


#------------------------------------------------------------------------------------------#
# Esta función devuelve los archivos de un directorio .mis limpios a un formato de archivo 
# especifico(.txt, .dat, .csv, ..., etc)
def cleanParsivel(PATH,PATH_OUT,formate):
    print('Limpiado archivos .mis ... \n ...')
    file_list = os.listdir(PATH)               # lista de archivos .mis contenidos en el directorio(carpeta)
    file_list.sort()                           # Ordenada alfabéticamente los ficheros del directorio
    for file in file_list:
        i = PATH + file                        # ruta del fichero
        print(i)
        j = PATH_OUT + file.replace(file[-4:],formate)
        cleanMIS(i,j)                            # Aplica función cleanMIS()
        print(j)
        print('Archivo limpio \n ...')
    print('\n PROCESO TERMINADO!')
        
        
        
        
#**************************************************************************************************#
# for i in ['12']:
#     PATH     = '/run/user/1000/gvfs/smb-share:server=synologynas.ugr.es,share=gfatnas/datos/parsivel/0a/2018/' + i +'/'
#     PATH_OUT = '/run/user/1000/gvfs/smb-share:server=synologynas.ugr.es,share=gfatnas/datos/parsivel/1a/2018/' + i +'/'
#     cleanParsivel(PATH,PATH_OUT,'.mis')
    


import numpy as np
from typing import Any


def HM_model_retrieval(
    Z_linear: np.ndarray,
    delta_h: float,
    model: str = "knist_N",
    lwp: list = None,
    nu: float = 8.7,
    sigma: float = 0.29,
    CDNC: float | None = None,
) -> list[np.ndarray, np.ndarray, np.ndarray]:
    """HM (Cloud Model): Radar Reflectivity-Homogeneous Mixing

    Args:
        Z_linear (np.ndarray): 1D array linear reflectivity
        delta_h (float): resolution in meters
        model (str, optional): approach used. Options: 'knist' | 'frisch'. Defaults to 'knist'.
        lwp (list, optional): Liquid Water Path, Scalar [Required for Knist''s approach]. Defaults to None.
        nu (float, optional): parameter to determine the shape of the log normal distrubion [Required for Knist''s approach]. Defaults to 8.7.
        sigma (float, optional): parameter to determine the width of the the log normal distrubion [Required for Frisch''s approach]. Defaults to 0.29.
        CDNC (float | None, optional): cloud-droplet number concenctration [cm^{-3}] [Required for Frisch''s approach]. Defaults to None.

    Returns:
        list[np.ndarray, np.ndarray, np.ndarray]: [re, N, Opt_Depth]
            - re: Effective radius. 1D array. (in um)
            - N: Droplet concentration. Scalar. (in cm-3). It is CDNC for Frisch''s approach
            - Opt_Depth: Optical Depth (Adimensional). It is empty for Frisch''s approach

    """
    if model == "knist_lwp":
        if len(lwp) > 0:
            WATER_DENSITY = 1e6  # g/m3
            kre = np.power(
                ((nu + 2) ** 3) / ((nu + 3) * (nu + 4) * (nu + 5)), 1.0 / 3.0
            )
            knt = ((nu + 3) * (nu + 4) * (nu + 5)) / (nu * (nu + 1) * (nu + 2))

            Z_linear_root = np.sqrt(Z_linear)
            sumZroot = Z_linear_root.sum(dim="height", skipna=True, min_count=1)
            # effective radius in um
            try:
                re = (
                    kre
                    * (Z_linear ** (1.0 / 6.0))
                    * np.power(
                        (np.pi * WATER_DENSITY * sumZroot * delta_h) / (48 * lwp),
                        1.0 / 3.0,
                    )
                )  # (Eq. 2.57 Knist PhD dissertation) Identical to Frisch et al. 2002 with sigma_x = 0.29 (value from marine Sc, Martin et al 1994)
                re *= 1e6
                # re = substitute_nan(re)
            except:
                re = np.empty(Z_linear.shape)
                re[:] = np.nan
            # droplet concentration in cm-3
            try:
                # N = knt*((6*lwp)/(np.pi*WATER_DENSITY*delta_h*np.nansum(np.sqrt(Z_linear))))**2 #Andrea's version
                N = knt * (
                    (6 * lwp) / np.power(np.pi * WATER_DENSITY * delta_h * sumZroot, 2)
                )
                N *= 1e-6
            except:
                N = np.nan
            # COD
            try:
                Opt_Depth = (
                    (3.0 / (2.0 * kre))
                    * ((48 / np.pi) ** (1.0 / 3.0))
                    * (lwp / np.power(WATER_DENSITY * delta_h * sumZroot), 4.0 / 3.0)
                    * delta_h
                    * np.power(sumZroot, 1.0 / 3.0)
                )
            except:
                Opt_Depth = np.nan
        else:
            print("ERROR: Knist" "s approximation requires lwp array.")
    elif model == "knist_N":
        if CDNC:
            WATER_DENSITY = 1e6  # g/m3
            kre = np.power(
                np.power(nu + 2, 5) / (nu * (nu + 1) * (nu + 3) * (nu + 4) * (nu + 5)),
                1.0 / 6.0,
            )
            # effective radius in um
            try:
                re = (
                    0.5 * kre * np.power(Z_linear / (CDNC * 1e6), 1.0 / 6.0)
                )  # (Eq. 2.56 Knist PhD dissertation) Identical to Frisch et al. 2002 with sigma_x = 0.29 (value from marine Sc, Martin et al 1994)
                re *= 1e6
            except:
                re = np.empty(Z_linear.shape)
                re[:] = np.nan
            # droplet concentration in cm-3
            N = CDNC * np.ones(len(re))
            # COD
            Opt_Depth = np.nan * np.ones(len(re))
        else:
            print("ERROR: Knist_N approximation requires CDNC.")
    elif model == "frisch":
        if CDNC:
            Opt_Depth = []
            # effective radius in um
            try:
                re = (
                    0.5
                    * np.exp(-0.5 * sigma**2)
                    * np.power(Z_linear / (CDNC * 1e6), 1.0 / 6.0)
                )  # (Frisch et al., 2002)
                # re = 0.5*np.power(Z_linear/(CDNC*1e6),1./6.) #(Frisch et al., 2002)
                # re = 0.5*np.exp(-0.5*sigma**2)*np.power(Z_linear/(CDNC*1e6),1./6.) #(Frisch et al., 2002)
                # re = 0.5*sigma*np.power(Z_linear/(CDNC*1e6),1./6.) #(Frisch et al., 2002)
                re *= 1e6
                # re = substitute_nan(re)
            except:
                re = np.empty(Z_linear.shape)
                re[:] = np.nan
            N = CDNC * np.ones(len(re))
        else:
            print("ERROR: Frisch" "s approximation requires lwp array.")
    return re, N, Opt_Depth


def LWC_model_HM(
    Z_linear: np.ndarray, LWP: float, delta_h: float
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """HM (Cloud Model): Radar Reflectivity-Homogeneous Mixing

    Args:
        Z_linear (np.ndarray): linear reflectivity.
        LWP (float):Liquid Water Path.
        delta_h (float): resolution in meters.

    Returns:
        np.ndarray[Any, np.dtype[np.float64]]: Liquid Water Content in g/m^2.
    """
    Z_linear_root = np.sqrt(Z_linear)
    sumZroot = Z_linear_root.sum(dim="height", skipna=True, min_count=1)
    LWC = LWP * (Z_linear_root / (sumZroot * delta_h))
    return LWC


def VU_model_retrieval(
    linear_Z: np.ndarray, lwp: np.ndarray, delta_H: np.ndarray, nu: float = 8.7
) -> list[np.ndarray]:
    """Vertical Uniform Model.

    Args:
        linear_Z (np.ndarray): linear reflectivity in m^3
        lwp (np.ndarray): Liquid Water Path in g·m^-2
        delta_H (np.ndarray): cloud thickness (m)
        nu (float, optional): droplet distribution parameter. Defaults to 8.7.

    Returns:
        list[np.ndarray]: [re, N, Opt_Depth]
            - re: Effective radius (in um)
            - N: Droplet concentration (in cm-3)
            - Opt_Depth: Optical Depth (Adimensional)
    """

    WATER_DENSITY = 1e6  # g/m3
    kre = ((nu + 2) ** 3 / ((nu + 3) * (nu + 4) * (nu + 5))) ** (1.0 / 3.0)
    knt = (nu + 3) * (nu + 4) * (nu + 5) / (nu * (nu + 1) * (nu + 2))

    # effective radius in um
    try:
        re = kre * ((np.pi * WATER_DENSITY * delta_H * linear_Z) / (48 * lwp)) ** (
            1.0 / 3.0
        )
        re *= 1e6
    except:
        re = np.nan

    # droplet concentration in cm-3
    try:
        N = (knt / linear_Z) * ((6 * lwp) / (np.pi * WATER_DENSITY * delta_H)) ** 2
        N *= 1e-6
    except:
        N = np.nan

    # optical depth (adimensional)
    try:
        Opt_Depth = (3.0 / 2.0) * (lwp / (WATER_DENSITY * (re / 1000000)))

    except:
        Opt_Depth = np.nan

    return re, N, Opt_Depth


def SAS_model_retrieval(
    Z_avg_linear: np.ndarray,
    lwp: float,
    height: np.ndarray,
    delta_H: float,
    nu: float = 8.7,
) -> list[np.ndarray, np.ndarray, np.ndarray]:
    """SCALED ADIABATIC STRATIFIED CLOUD MODEL

    Args:
        Z_avg_linear (np.ndarray): linear) Z in cloud for each profile
        lwp (float): liquid water path in g·m^-2
        height (np.ndarray): height in m.
        delta_H (float): cloud thickness (m)
        nu (float, optional): droplet distribution parameter. Defaults to 8.7.

    Returns:
        list[np.ndarray, np.ndarray, np.ndarray]: [re, N, Opt_Depth]
            - re: Effective radius (in um)
            - N: Droplet concentration (in cm-3)
            - Opt_Depth: Optical Depth (Adimensional)
    """

    WATER_DENSITY = 1000000  # g/m3
    kre = ((nu + 2) ** 3 / ((nu + 3) * (nu + 4) * (nu + 5))) ** (1.0 / 3.0)
    knt = (nu + 3) * (nu + 4) * (nu + 5) / (nu * (nu + 1) * (nu + 2))

    # effective radius in um

    try:
        re = (
            height ** (1.0 / 3.0)
            * kre
            * ((np.pi * WATER_DENSITY * Z_avg_linear) / (32 * lwp)) ** (1.0 / 3.0)
        )  # um
        re *= 1e6
    except:
        re = np.nan

    # Maximum radius
    re_max = np.nanmax(re)

    # Optical_Depth (adimensional)
    try:
        Opt_Depth = (9.0 / 5.0) * (lwp / (WATER_DENSITY * (re_max / 1000000)))
    except:
        Opt_Depth = np.nan

    # droplet concentration in cm-3
    try:
        N = (
            (4.0 / 3.0)
            * (knt / Z_avg_linear)
            * ((6 * lwp) / (np.pi * WATER_DENSITY * delta_H)) ** 2
        )
        N *= 1e-6
    except:
        N = np.nan

    return re, N, Opt_Depth



from pathlib import Path


from datetime import datetime
import numpy as np

from gfatpy.lidar.utils.utils import LIDAR_INFO
from gfatpy.utils.utils import parse_datetime


def channel2info(channel: str) -> tuple[int, str, str | None, str]:
    """`channel_str2info` extracts the information concerning the channel configuration from the string code.

    Args:
        channel (str): Chanel string code (e.g., '532xta'.)

    Returns:
        tuple[str]: Tuple with following elements: wavelength [int], telescope Options['xf', 'ff', 'nf'], polarization Options['p', 's', 't'], detection Options['a', 'p', 'g'].
    """

    channel = channel.split("_")[-1]
    if len(channel) > 5:
        wave = int(channel[0:-3], base=10)
        telescope = f"{channel[-3]}f"
        polarization = channel[-2]
        detection = channel[-1]
    elif len(channel) == 5:
        wave = int(channel[0:-2], base=10)
        telescope = f"{channel[-2]}f"
        polarization = None
        detection = channel[-1]
    else:
        raise ValueError("Channel format not recognized.")

    return wave, telescope, polarization, detection


def filename2info(filename: str) -> tuple[str, str, str, str, str, datetime]:
    parts = filename.split("_")

    lidar_nick = parts[0]
    data_level = parts[1]
    measurement_type = parts[2][1:3]
    signal_type = parts[3]
    telescope = parts[4]

    if measurement_type in ["tc", "dp"]:  # Fallback new format nc for TC and DP
        signal_type = parts[2][1:3]
        telescope = parts[3]
        date_ = parts[4].split(".")[0]
    else:
        date_ = parts[5].split(".")[0]

    if len(parts) == 7:  # has hour
        hour_ = parts[6].split(".")[0]
        date = datetime.strptime(date_ + hour_, "%Y%m%d%H%M")
    elif measurement_type in ["tc", "dp"]:
        hour_ = parts[5].split(".")[0]
        date = datetime.strptime(date_ + hour_, "%Y%m%d%H%M")
    else:  # has not hour
        date = datetime.strptime(date_, "%Y%m%d")

    return lidar_nick, data_level, measurement_type, signal_type, telescope, date


def info2filename(
    channel: str,
    date: datetime,
    lidar_nick: str | None = None,
    data_level: str = "1a",
    measurement_type: str = "RS",
    signal_type: str = "rs",
    lidar_name: str | None = None,
) -> str:
    """Lidar information to filename.

    Args:
        channel (str): Channel string code (e.g., '532xpa')
        date (datetime.datetime | datetime.date): It must be datetime.datetime if `measurement_type` is `DC`.
        lidar_nick (str | None): Lidar nick (e.g., 'mhc'0')
        measurement_type (str, optional): Measurment lidar type ['RS', 'DP', 'TC', DC', 'OT']. Defaults to 'RS'.
        signal_type (str, optional): Signal type  ['rs', 'sd']. Defaults to 'rs'.
        lidar_name (str | None, optional): If `lidar_nick` is not provided, `lidar_name` is mandatory. Defaults to None.

    Returns:
        str: filename
    """

    if lidar_nick is None and lidar_name is None:
        raise NameError("lidar_nick or lidar_name must be provided.")

    if lidar_nick is None and lidar_name is not None:
        if lidar_name in LIDAR_INFO["metadata"]["name2nick"].keys():
            lidar_nick = LIDAR_INFO["metadata"]["name2nick"][lidar_name]

    if lidar_nick in LIDAR_INFO["metadata"]["nick2name"].keys():
        _, telescope, _, _ = channel2info(channel)
    else:
        raise NameError("lidar_nick not found in `LIDAR_INFO`.")

    date_str = date.strftime("%Y%m%d")

    if measurement_type not in ["TC", "DP"]:
        filename_ = f"{lidar_nick}_{data_level}_P{measurement_type.lower()}_{signal_type.lower()}_{telescope}_{date_str}"
    else:
        filename_ = f"{lidar_nick}_{data_level}_P{measurement_type.lower()}-{signal_type}_{telescope}_{date_str}"

    if measurement_type in ["DC", "TC", "DP", "OT"]:
        if isinstance(date, datetime):
            hour = date.strftime("%H%M")
            filename_ = filename_ + "_" + hour
        else:
            raise TypeError(
                "Date must be datetime.datetime if `measurement_type` is `DC`."
            )

    filename = filename_ + ".nc"

    return filename


def filename2path(
    filename: str,
    dir: Path,
    data_level: str | None = None,
    check_exists: bool = True,
) -> Path:
    """Create the path where is located the `filename`.

    Args:
        filename (str): lidar file name (e.g, 'mhc_1a_Prs_rs_xf_20220808.nc')
        dir (Path): Root directory where lidar data are located.
        check_exist (bool, optional): Force to check if the file exists in that path. If False, it raises an error. Defaults to True.

    Returns:
        filepath (Path): lidar file path.
    """

    if not dir.exists() or not dir.is_dir():
        raise ValueError("Path must be provided")

    lidar_nick, _, *_, date = filename2info(filename)

    lidar_name: str = LIDAR_INFO["metadata"]["nick2name"][lidar_nick]

    if data_level is None:
        filepath = (
            dir
            / lidar_name
            / f"{date.year}"
            / f"{date.month:02d}"
            / f"{date.day:02d}"
            / filename
        )
    else:
        filepath = (
            dir
            / lidar_name
            / data_level
            / f"{date.year}"
            / f"{date.month:02d}"
            / f"{date.day:02d}"
            / filename
        )

    if check_exists and not filepath.is_file():
        raise FileNotFoundError(f"{filepath} does not exist.")

    return filepath


def info2path(
    lidar_name: str,
    channel: str,
    date: datetime,
    dir: Path,
    data_level: str = "1a",
    measurement_type: str = "RS",
    signal_type: str = "rs",
    check_exist: bool = False,
) -> Path:
    """`directory_from_info` provides the directory of 1a-lidar files.

    Args:
        lidar_name (str): Lidar name. Options in `LIDAR_INFO`.
        channel (str): channel code. Options in `LIDAR_INFO`.
        date (datetime.date): Date of measurement.
        dir (Path): Root where lidar data are storaged.
        data_level (str, optional): Data level. Defaults to '1a'.

    Raises:
        NotFoundErr: Directory created does not exist.

    Returns:
        Path: Directory with 1a-lidar files.
    """

    if not dir.exists() or not dir.is_dir():
        raise NotADirectoryError("dir not found.")

    filename = info2filename(
        channel,
        date,
        data_level=data_level,
        lidar_name=lidar_name,
        measurement_type=measurement_type,
        signal_type=signal_type,
    )

    filepath = filename2path(
        filename, dir, data_level=data_level, check_exists=check_exist
    )

    return filepath


def search_dc(
    rs_path: Path,
    session_period: tuple[np.datetime64, np.datetime64] | np.ndarray,
    force_dc_in_session: bool = False,
) -> Path:
    # TODO: Change this documentation
    # TODO: Is fine only searching in the same day?
    """`find_dc` searches a DC file path near to the `date` provided that can be used to preprocess the `lidar_data` data.

    Args:
        lidar_nick (str): Nick of lidar. Options in `LIDAR_INFO`.
        telescope (str, optional): Telescope type code Options['xf', 'ff', 'nf']
        session_period (tuple[datetime.datetime, datetime.datetime]): Period where the DC is searched. If not found, the nearest is provided.
        force_dc_in_session (bool, optional): Force the DC measurment to be in the period, otherwise it raises an error. Defaults to False.
    Returns:
        Path: Path of the DC measurement.
    """
    telescope = filename2info(rs_path.name)[4]
    candidates = list(rs_path.parent.glob(f"*1a_Pdc_*{telescope}*"))

    for dc_path in candidates:
        dc_npdate = np.datetime64(filename2info(dc_path.name)[-1])
        if (session_period[0] <= dc_npdate) and (dc_npdate <= session_period[1]):
            return dc_path

    if force_dc_in_session:
        raise FileNotFoundError(f"Cannot find a dc for times {session_period}")

    candidates, candidate_dates = extract_filenames_dates_from_wildcard(
        rs_path.parent.parent, "*1a_Pdc_*"
    )
    idx = np.abs(
        candidate_dates.astype("M8[ns]").astype(float)
        - np.array(session_period).astype("M8[ns]").astype(float).mean()
    ).argmin()
    dc_path = candidates[idx]

    return dc_path


def extract_filenames_dates_from_wildcard(
    directory: Path, file_wildcard: str
) -> tuple[list[Path], np.ndarray]:
    """Provides file list and its dates from file wildcard. WARNING: slow function.

    Args:
        directory (Path): Directory where files will be searched, including subdirectories.
        file_wildcard (str): string file wildcard (e.g., '*Prs*.nc')

    Raises:
        FileNotFoundError: Any file found.

    Returns:

        np.ndarray: array of dates.
    """

    candidates = list(filter(lambda p: p.is_file(), directory.rglob(file_wildcard)))

    if len(candidates) == 0:
        raise FileNotFoundError(
            f"File not found with description {file_wildcard} in {directory}."
        )

    dates = np.array(
        list(
            map(
                lambda p: np.datetime64(p.name.split(".")[0].split("_")[-1]), candidates
            )
        )
    )
    return candidates, dates


def info2general_path(
    lidar_name: str,
    date: datetime | str,
    data_dir: Path,
    data_level: str | None = None,
) -> Path:
    """Similar to info2path but can take only lidar name, data level and dates as only args

    Args:
        lidar_name (str): _description_
        date (datetime | str): _description_
        data_dir (Path): Path up to lidar folders.
        data_level (str): _description_

    Returns:
        Path: _description_
    """

    date = parse_datetime(date)

    if not data_dir.exists() or not data_dir.is_dir():
        raise NotADirectoryError("data_dir not found.")

    if (data_level is not None) and (data_level != "0a"):
        return (
            data_dir
            / lidar_name
            / data_level
            / f"{date.year}"
            / f"{date.month:02}"
            / f"{date.day:02}"
        )
    else:
        return (
            data_dir
            / lidar_name
            / f"{date.year}"
            / f"{date.month:02}"
            / f"{date.day:02}"
        )



from pathlib import Path

import numpy as np
import xarray as xr
from datetime import datetime

from gfatpy.lidar.utils import file_manager

from gfatpy.utils.io import find_nearest_filepath


def search_nereast_calib(
    depoCalib: dict, channel: str, current_date: np.datetime64
) -> dict:
    """It searches the nearest calibration to `current_date`. 

    Args:
        depoCalib (dict): Dictionary with depolarization calibrations.
        channel (str): Target channel.
        current_date (np.datetime64): Datetime.

    Returns:
        dict: Dictionary with depolarization calibration at target datetime. 
    """
    wavelength_, telescope_, *_ = file_manager.channel2info(channel)

    # Search the last calibration performed the current measurement
    idx = depoCalib[telescope_]["%0d" % wavelength_].index.get_indexer(
        [current_date], method="pad"
    )  # 'pad': search the nearest lower; 'nearest': search the absolute nearest.

    calib = depoCalib[telescope_]["%0d" % wavelength_].iloc[idx].to_dict("records")[0]

    return calib


def search_nearest_eta_star_from_file(
    channel: str, target_date: datetime, calib_dir: Path | str | None = None
) -> tuple[float, float]:
    """It searches the nearest calibration factor file to a given date (`target_date`).

    Args:
        channel (str): target channel.
        target_date (datetime): target datetime.
        calib_dir (Path | str | None, optional): Directory of calibrations. Defaults to None.

    Raises:
        NotADirectoryError: `calib_dir` is not a directory.

    Returns:
        tuple[float, float]: depolarization calibration average and standard deviation.
    """    
    if calib_dir is None:
        calib_dir = Path.cwd()
    elif isinstance(calib_dir, str):
        calib_dir = Path(calib_dir)

    if not calib_dir.exists() or not calib_dir.is_dir():
        raise NotADirectoryError(f"{calib_dir} not found.")

    calib_path = find_nearest_filepath(calib_dir, "*eta-star*", 2, target_date, and_previous=True)

    calib = xr.open_dataset(calib_path)

    eta_star_mean = f"eta_star_mean_{channel}"
    et_star = calib[eta_star_mean].values.item()
    std_et_star = calib[f"std_{eta_star_mean}"].values.item()

    return et_star, std_et_star



import datetime as dt
from pathlib import Path
from pdb import set_trace
import numpy as np
import xarray as xr

from gfatpy.utils.utils import numpy_to_datetime
from gfatpy.lidar.preprocessing.lidar_preprocessing import preprocess
from gfatpy.lidar.utils.file_manager import channel2info


def retrieve_ff_overlap(
    filepath: Path | str,
    hour_range: tuple[float, float],
    output_dir: Path,
    norm_range: tuple[float, float] = (2500, 3500),
    rel_dif_threshold: float = 2.5,
) -> Path | None:
    """Retrieve ff overlap from a near-to-far module lidar ratio

    Args:

        - filepath (Path | str): Lidar file path
        - hour_range (tuple[float, float]): Hour range to calculate the overlap
        - norm_range (tuple[float, float], optional): Range to normalized signals. Defaults to (2500, 3500).
        - rel_dif_threshold (float, optional): Relative difference threshold. Overlap function will be rejected if it causes a relative difference larger than this threshold. Defaults to 2.5 %.
        - output_dir (Path, optional): output folder path. Defaults to Path.cwd().

    Raises:

        - FileNotFoundError: Lidar file not found
        - ValueError: Hour range must be a tuple of two floats
        - ValueError: Not enough profiles in file to cover the selected time range [<50%]
        - ValueError: Could not calculate overlap for 355 nm
        - ValueError: Could not calculate overlap for 532 nm
        - ValueError: Could not calculate overlap for 1064 nm
        - ValueError: Could not calculate overlap for any channel

    Returns:

        - Path | None : Path to the overlap file
    """

    if isinstance(filepath, str):
        filepath = Path(filepath)

    if not filepath.exists():
        raise FileNotFoundError(f"File {filepath} not found")

    if hour_range[0] > hour_range[1]:
        raise ValueError("hour_range[0] must be lower than hour_range[1]")

    # Check time period is within file time range
    raw = xr.open_dataset(filepath)

    # Get date from filename
    date_ = numpy_to_datetime(raw.time[2].values)
    if date_ is None:
        raise ValueError("Could not get date from filename")
    date = dt.datetime(date_.year, date_.month, date_.day)

    time_range = (
        (date + dt.timedelta(hours=hour_range[0])),
        (date + dt.timedelta(hours=hour_range[-1])),
    )
    time_resol = (raw.time[1] - raw.time[0]).item() / 1e9
    expected_profile_number = (
        time_range[1] - time_range[0]
    ).total_seconds() / time_resol
        
    if raw.sel(time=slice(*time_range)).time.size < 0.5 * expected_profile_number:
        raise ValueError(
            "Not enough profiles in file to cover the selected time range [<50%]"
        )

    raw.close()

    # Preprocess lidar file
    lidar_ = preprocess(
        filepath
    )  # FIXME: add crop_ranges and gluing_products and apply_dt as arguments

    # Select time range
    lidar = lidar_.sel(time=slice(*time_range))

    # Retrieve overlap at 355, 532, 1064 nm
    channels = []
    try:
        overlap355 = ff_overlap_from_channels( lidar, "355fpa", "355npa", norm_range=norm_range, rel_dif_threshold=rel_dif_threshold, )
        channels.append("355fpa")
    except:
        overlap355 = None
        raise ValueError("Could not calculate overlap for 355 nm")
    try:
        overlap532 = ff_overlap_from_channels(
            lidar,
            "532fta",
            "532npa",
            norm_range=norm_range,
            rel_dif_threshold=rel_dif_threshold,
        )
        channels.append("532fta")
    except:
        overlap532 = None
        raise ValueError("Could not calculate overlap for 532 nm")
    try:
        overlap1064 = ff_overlap_from_channels(
            lidar,
            "1064fta",
            "1064nta",
            norm_range=norm_range,
            rel_dif_threshold=rel_dif_threshold,
        )
        channels.append("1064fta")
    except:
        overlap1064 = None
        raise ValueError("Could not calculate overlap for 1064 nm")

    # Merge overlap data

    overlap_matrix = np.zeros((len(lidar.range), len(channels)))
    if overlap355 is not None and overlap532 is not None and overlap1064 is not None:
        overlap_matrix[:, 0] = overlap355.values
        overlap_matrix[:, 1] = overlap532.values
        overlap_matrix[:, 2] = overlap1064.values
    elif overlap355 is not None and overlap532 is not None:
        overlap_matrix[:, 0] = overlap355.values
        overlap_matrix[:, 1] = overlap532.values
    elif overlap355 is not None and overlap1064 is not None:
        overlap_matrix[:, 0] = overlap355.values
        overlap_matrix[:, 2] = overlap1064.values
    elif overlap532 is not None and overlap1064 is not None:
        overlap_matrix[:, 1] = overlap532.values
        overlap_matrix[:, 2] = overlap1064.values
    elif overlap355 is not None:
        overlap_matrix[:, 0] = overlap355.values
    elif overlap532 is not None:
        overlap_matrix[:, 1] = overlap532.values
    elif overlap1064 is not None:
        overlap_matrix[:, 2] = overlap1064.values
    else:
        overlap_matrix = None
        raise ValueError("Could not calculate overlap for any channel")

    if overlap_matrix is not None:
        # Create DataArray
        overlap = xr.DataArray(
            overlap_matrix,
            dims=("range", "channel"),
            coords={"range": lidar.range.values, "channel": channels},
        )

        # Create output folder
        if not output_dir.exists() and not output_dir.is_dir():
            output_dir.mkdir(parents=True)

        time_min = numpy_to_datetime(lidar.time[0].values)
        if time_min is None:
            raise ValueError("Could not get time from filename")
        time_min = time_min.strftime("%H%M")

        time_max = numpy_to_datetime(lidar.time[-1].values)
        if time_max is None:
            raise ValueError("Could not get time from filename")
        time_max = time_max.strftime("%H%M")

        # Create output filename
        if not output_dir.exists() or not output_dir.is_dir():
            raise FileNotFoundError(f"Output directory {output_dir} does not exist")

        output_path = (
            output_dir
            / f'overlap_alh_ff_{date.strftime("%Y%m%d")}_{time_min}-{time_max}.nc'
        )

        # save overlap files
        overlap.to_netcdf(output_path)
        print(f"Overlap saved in {output_path}")
    else:
        raise ValueError("Could not calculate overlap for any channel")
    return output_path


def ff_overlap_from_channels(
    lidar_dataset: xr.Dataset,
    channel_ff: str,
    channel_nf: str,
    norm_range: tuple[float, float] = (2500, 3500),
    rel_dif_threshold: float = 2.5,
) -> xr.DataArray:

    info_ff = channel2info(channel_ff)
    info_nf = channel2info(channel_nf)

    if info_nf[0] != info_ff[0]:
        raise ValueError(
            f"Channels {channel_ff} and {channel_nf} must have the same wavelength"
        )
    
    #check if lidar_dataset is a dask array
    is_dask = any(var.chunks is not None for var in lidar_dataset.variables.values())
    if is_dask:
        lidar_dataset = lidar_dataset.compute()  # This line computes all Dask arrays to numpy arrays

    # Select time, range to normalize and range to calculate overlap
    nf = lidar_dataset[f"signal_{channel_nf}"].mean("time") / lidar_dataset[
        f"signal_{channel_nf}"
    ].mean("time").sel(range=slice(*norm_range)).mean("range")
    ff = lidar_dataset[f"signal_{channel_ff}"].mean("time") / lidar_dataset[
        f"signal_{channel_ff}"
    ].mean("time").sel(range=slice(*norm_range)).mean("range")

    # Overlap
    overlap_raw = ff / nf

    # Upper limit
    overlap = overlap_raw.copy()
    median = overlap.rolling(range=11, center=True).median("range")
    max_overlap = overlap.range[1 - median.values < 0][0]
    overlap[overlap.range > max_overlap] = 1.0

    # Lower limit
    rel_dif = 100 * (nf - ff / overlap) / nf
    overlap[(overlap.range < 100)] = np.nan
    overlap[rel_dif > rel_dif_threshold] = np.nan

    # Upper limit
    overlap[overlap.range > max_overlap] = 1.0

    # Assign attributes
    attrs = ["location", "system"]
    for attr_ in attrs:
        overlap.attrs[attr_] = lidar_dataset.attrs[attr_]

    overlap.attrs["history"] = dt.datetime.now().strftime(
        "Created %a %b %d %H:%M:%S %Y"
    )
    overlap.attrs["wavelength"] = info_nf[0]
    overlap.attrs["channel_ff"] = channel_ff
    overlap.attrs["channel_nf"] = channel_nf

    return overlap



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gfatpy',
 'gfatpy.aeronet',
 'gfatpy.atmo',
 'gfatpy.cli',
 'gfatpy.cli.lidar',
 'gfatpy.cli.lidar.plot',
 'gfatpy.cloudnet',
 'gfatpy.generalife',
 'gfatpy.generalife.licel',
 'gfatpy.lidar',
 'gfatpy.lidar.depolarization',
 'gfatpy.lidar.depolarization.GHK',
 'gfatpy.lidar.depolarization.GHK.system_settings',
 'gfatpy.lidar.depolarization.deprecated',
 'gfatpy.lidar.nc_convert',
 'gfatpy.lidar.plot',
 'gfatpy.lidar.preprocessing',
 'gfatpy.lidar.quality_assurance',
 'gfatpy.lidar.retrieval',
 'gfatpy.lidar.retrieval.synthetic',
 'gfatpy.lidar.scc',
 'gfatpy.lidar.scc.licel2scc',
 'gfatpy.lidar.scc.licel2scc.systems',
 'gfatpy.lidar.scc.plot',
 'gfatpy.lidar.scc.scc_configFiles',
 'gfatpy.lidar.utils',
 'gfatpy.parsivel',
 'gfatpy.parsivel.michi_to_convert',
 'gfatpy.parsivel.plot',
 'gfatpy.radar',
 'gfatpy.radar.leonie von Terzi',
 'gfatpy.radar.plot',
 'gfatpy.radar.retrieve',
 'gfatpy.utils',
 'gfatpy.worker']

package_data = \
{'': ['*'],
 'gfatpy': ['assets/*', 'env_files/*'],
 'gfatpy.lidar': ['info/*', 'nc_convert/configs/*', 'scc/scc_campaigns/*'],
 'gfatpy.parsivel': ['scattering_databases/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'atmospheric_lidar>=0.5.4,<0.6.0',
 'dask>=2022.7.1,<2023.0.0',
 'linc>=1.8.0,<2.0.0',
 'loguru>=0.6.0,<0.7.0',
 'matplotlib>=3.5.2,<4.0.0',
 'netCDF4>=1.6.2,<2.0.0',
 'numba>=0.56.2,<0.57.0',
 'numpy>=1.23.1,<2.0.0',
 'pyarrow>=10.0.1,<11.0.0',
 'requests>=2.28.1,<3.0.0',
 'rpgpy>=0.15.1,<0.16.0',
 'scikit-image>=0.19.3,<0.20.0',
 'scikit-learn>=1.1.1,<2.0.0',
 'scipy>=1.9.0,<2.0.0',
 'seaborn>=0.12.0,<0.13.0',
 'selenium>=4.1.0,<5.0.0',
 'statsmodels>=0.13.0,<0.14.0',
 'typer[all]>=0.6.1,<0.7.0',
 'typing-extensions>=4.3.0,<5.0.0',
 'xarray[h5netcdf]>=2023.6.0,<2024.0.0']

extras_require = \
{'docs': ['pdoc>=12.0.2,<13.0.0']}

entry_points = \
{'console_scripts': ['gfatpy = gfatpy.cli.main:app']}

setup_kwargs = {
    'name': 'gfatpy',
    'version': '0.14.4',
    'description': 'A python package for GFAT utilities',
    'long_description': '# gfatpy\n\nA python package for GFAT utilities\n\n\n[![pipeline status](https://gitlab.com/gfat1/gfatpy/badges/main/pipeline.svg)](https://gitlab.com/gfat1/gfatpy/-/commits/main)\n\n\n---\n**Documentation** : [https://gfat1.gitlab.io/gfatpy](https://gfat1.gitlab.io/gfatpy/gfatpy.html)  \n**Source Code** : [https://gitlab.com/gfat1/gfatpy](https://gitlab.com/gfat1/gfatpy)  \n\n---\n\n## Create package\nFollowing guide at: https://packaging.python.org/en/latest/tutorials/packaging-projects/#classifiers\n\nin gfatpy directory, run in terminal: python -m build\nIt creates "dist" directory .whl and tar.gz files\n\n## Install package in a virtual environment\npip install [path]/dist/gfatpy-0.0.0.tar.gz\nNB: install pip: conda install pip\nNB: install python: conda install ipython\n\n# gitlab stuff\n## Getting started\n\nTo make it easy for you to get started with GitLab, here\'s a list of recommended next steps.\n\nAlready a pro? Just edit this README.md and make it your own. Want to make it easy? [Use the template at the bottom](#editing-this-readme)!\n\n## Add your files\n\n- [ ] [Create](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/user/project/repository/web_editor.html#create-a-file) or [upload](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/user/project/repository/web_editor.html#upload-a-file) files\n- [ ] [Add files using the command line](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/gitlab-basics/add-file.html#add-a-file-using-the-command-line) or push an existing Git repository with the following command:\n\n```\ncd existing_repo\ngit remote add origin https://gitlab.com/gfat1/gfatpy.git\ngit branch -M main\ngit push -uf origin main\n```\n\n## Integrate with your tools\n\n- [ ] [Set up project integrations](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://gitlab.com/gfat1/gfatpy/-/settings/integrations)\n\n## Collaborate with your team\n\n- [ ] [Invite team members and collaborators](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/user/project/members/)\n- [ ] [Create a new merge request](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/user/project/merge_requests/creating_merge_requests.html)\n- [ ] [Automatically close issues from merge requests](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/user/project/issues/managing_issues.html#closing-issues-automatically)\n- [ ] [Enable merge request approvals](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/user/project/merge_requests/approvals/)\n- [ ] [Automatically merge when pipeline succeeds](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/user/project/merge_requests/merge_when_pipeline_succeeds.html)\n\n## Test and Deploy\n\nUse the built-in continuous integration in GitLab.\n\n- [ ] [Get started with GitLab CI/CD](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/ci/quick_start/index.html)\n- [ ] [Analyze your code for known vulnerabilities with Static Application Security Testing(SAST)](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/user/application_security/sast/)\n- [ ] [Deploy to Kubernetes, Amazon EC2, or Amazon ECS using Auto Deploy](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/topics/autodevops/requirements.html)\n- [ ] [Use pull-based deployments for improved Kubernetes management](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/user/clusters/agent/)\n- [ ] [Set up protected environments](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://docs.gitlab.com/ee/ci/environments/protected_environments.html)\n\n***\n\n# Editing this README\n\nWhen you\'re ready to make this README your own, just edit this file and use the handy template below (or feel free to structure it however you want - this is just a starting point!).  Thank you to [makeareadme.com](https://gitlab.com/-/experiment/new_project_readme_content:3a94b1732a94fc66f6509822a9d9c9d8?https://www.makeareadme.com/) for this template.\n\n## Suggestions for a good README\nEvery project is different, so consider which of these sections apply to yours. The sections used in the template are suggestions for most open source projects. Also keep in mind that while a README can be too long and detailed, too long is better than too short. If you think your README is too long, consider utilizing another form of documentation rather than cutting out information.\n\n## Name\nChoose a self-explaining name for your project.\n\n## Description\nLet people know what your project can do specifically. Provide context and add a link to any reference visitors might be unfamiliar with. A list of Features or a Background subsection can also be added here. If there are alternatives to your project, this is a good place to list differentiating factors.\n\n## Badges\nOn some READMEs, you may see small images that convey metadata, such as whether or not all the tests are passing for the project. You can use Shields to add some to your README. Many services also have instructions for adding a badge.\n\n## Visuals\nDepending on what you are making, it can be a good idea to include screenshots or even a video (you\'ll frequently see GIFs rather than actual videos). Tools like ttygif can help, but check out Asciinema for a more sophisticated method.\n\n## Installation\nWithin a particular ecosystem, there may be a common way of installing things, such as using Yarn, NuGet, or Homebrew. However, consider the possibility that whoever is reading your README is a novice and would like more guidance. Listing specific steps helps remove ambiguity and gets people to using your project as quickly as possible. If it only runs in a specific context like a particular programming language version or operating system or has dependencies that have to be installed manually, also add a Requirements subsection.\n\n## Usage\nUse examples liberally, and show the expected output if you can. It\'s helpful to have inline the smallest example of usage that you can demonstrate, while providing links to more sophisticated examples if they are too long to reasonably include in the README.\n\n## Support\nTell people where they can go to for help. It can be any combination of an issue tracker, a chat room, an email address, etc.\n\n## Roadmap\nIf you have ideas for releases in the future, it is a good idea to list them in the README.\n\n## Contributing\nState if you are open to contributions and what your requirements are for accepting them.\n\nFor people who want to make changes to your project, it\'s helpful to have some documentation on how to get started. Perhaps there is a script that they should run or some environment variables that they need to set. Make these steps explicit. These instructions could also be useful to your future self.\n\nYou can also document commands to lint the code or run tests. These steps help to ensure high code quality and reduce the likelihood that the changes inadvertently break something. Having instructions for running tests is especially helpful if it requires external setup, such as starting a Selenium server for testing in a browser.\n\n## Authors and acknowledgment\nShow your appreciation to those who have contributed to the project.\n\n## License\nFor open source projects, say how it is licensed.\n\n## Project status\nIf you have run out of energy or time for your project, put a note at the top of the README saying that development has slowed down or stopped completely. Someone may choose to fork your project or volunteer to step in as a maintainer or owner, allowing your project to keep going. You can also make an explicit request for maintainers.\n',
    'author': 'Juan Diego De la Rosa',
    'author_email': 'jdidelarc@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gfat1.gitlab.io/gfatpy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)




from pathlib import Path
import numpy as np
import pandas as pd
import xarray as xr
from datetime import datetime

from gfatpy.lidar.nc_convert.types import Measurement
from gfatpy.generalife.licel import LicelReader

DZ = 7.5
MAX_BINS = 2000

def licel_to_datetime(licel_name: str) -> datetime | None:
    name, extension = licel_name.split(".")
    try:
        month_decimal = int(name[-5], 16)
        return datetime.strptime(f"{name[-7:-5]}{month_decimal:02d}{name[-4:-2]}T{name[-2:]}{extension[:4]}", r"%y%m%dT%H%M%S")
    except ValueError:
        print(f"Skipping file {licel_name} due to unexpected name format.")
        return None
    

def measurement_to_nc(measurement: Measurement, output_dir: Path):    
    files = measurement.get_filepaths()
    
    times = [licel_to_datetime(file.name) for file in files]
    
    df = pd.read_csv([*files][0].absolute().as_posix(), header=None, skiprows=3, delimiter=' ', encoding='ISO-8859-1', on_bad_lines='skip')
    wavelengths = df[8].dropna().astype(float).values
    voltages = df[6].dropna().astype(float).values

    data = np.empty((len(files), len(wavelengths), MAX_BINS), dtype=float)

    for file_idx, file_path in enumerate(files):
        sp = LicelReader.LicelFileReader(file_path)
        for i, _ in enumerate(wavelengths):
            data[file_idx, i, :] = sp.dataSet[i].physData

    # Create an xarray DataArray from the data array
    raw = xr.DataArray(
        data,
        coords={'wavelength': wavelengths, 'times': times, 'range': DZ*np.arange(MAX_BINS)},
        dims=['times', 'wavelength', 'range']
    )

    voltages = xr.DataArray(
        voltages,
        coords={'wavelength': wavelengths},
        dims=['wavelength']
    )

    center_wavelength = (wavelengths[16] + wavelengths[15]) / 2

    # Create an xarray Dataset from the DataArray
    ds = xr.Dataset({'raw': raw, 'voltages': voltages})

    ds.attrs.update({
            'system name': 'Generalife',  
            'location': 'Granada',
            'latitude': '37.1641° N',
            'longitude': '3.6026° W',
            'altitude': '680 m',
            'frequency': '10 Hz',
            'bin width': '7.5 m',
            'number of bins': '2000',
            'number of channels': '32',
            'center wavelength': center_wavelength,            
            'discriminator' : '4.0',
            'bin_corrected': 'False',
            'bcg_corrected': 'False',
        })

    if isinstance(times[0], datetime):
        datestr = times[0].strftime('%Y%m%d')   
        filename = f"gnl_1a_Prs_rs_ff_{datestr}.nc"
        output_path = output_dir / "generalife" / "1a" / f"{times[0].strftime('%Y')}" / f"{times[0].strftime('%m')}" / f"{times[0].strftime('%d')}" / filename
        ds.to_netcdf(output_path)




import warnings
from typing import Any
from datetime import datetime, timedelta
from functools import cached_property

import numpy as np
import xarray as xr
import matplotlib.pyplot as plt


from .klett import klett_rcs
from gfatpy.atmo.freudenthaler_molecular_properties import molecular_properties
from gfatpy.utils.utils import parse_datetime
from gfatpy.lidar.plot.quicklook import quicklook_xarray
from gfatpy.atmo.ecmwf import get_ecmwf_temperature_pressure
from gfatpy.lidar.plot.profile import plot_profile


warnings.filterwarnings("once")


class RetrievalHelper:
    def __init__(self, preprocessed_data: xr.Dataset, channel: str = "532xpa"):
        if channel in preprocessed_data.channel:
            self._data = preprocessed_data
            self._channel = channel
        else:
            raise ValueError(
                f"Channel {channel} not available in current dataset select another one from: {preprocessed_data.channel}"
            )

        self._k_lidar: float | None = None

    @cached_property
    def wavelength(self) -> int:
        return int("".join(filter(str.isdigit, self.channel)))

    @property
    def channel(self) -> str:
        return self._channel

    @property
    def data(self) -> xr.Dataset:
        return self._data

    @property
    def k_lidar(self) -> float:
        if self._k_lidar is None:
            raise ValueError("k_lidar has not been calibrated")
        return self._k_lidar

    @cached_property
    def rcs(self) -> np.ndarray:
        return self.data[f"signal_{self.channel}"].values * np.tile(
            self.data.range**2, (self.data.time.shape[0], 1)
        )

    def closest_profile(self, date: datetime | str, print_date=False) -> int:
        _date = parse_datetime(date)
        idx = np.searchsorted(self.data.time, np.datetime64(_date))

        date_found = (
            self.data.time[idx].values.astype("M8[ms]").astype("O")
        )  # This is to convert to pydatetime

        distance = date_found - _date

        td = timedelta(minutes=60)

        if (distance > td) or (distance < -td):
            warnings.warn(
                f"Using datetime with more than 1 hour of difference from the input ({date_found})"
            )

        if print_date:
            print(f"Date found: {date_found} at {distance} distance")

        return idx  # type: ignore

    def plot_profile(self, profile_idx: int, rcs: bool = True):
        if rcs:
            plot_profile(self.data.range, self.rcs[profile_idx])
        else:
            raise NotImplementedError()

    def klett_on_profile(self, profile_idx: int) -> Any:
        date: datetime = self.data.time.values.astype("M8[ms]").astype("O")[profile_idx]
        tp_df = get_ecmwf_temperature_pressure(date, heights=self.data.range.values)
        atmo_params = molecular_properties(
            self.wavelength,
            tp_df.pressure,
            tp_df.temperature,
            heights=self.data.range.values,
        )
        # set_trace()
        part_beta = klett_rcs(self.rcs[profile_idx], range_profile=self.data.range.values, beta_mol_profile=atmo_params["molecular_beta"].values)  # type: ignore

        return part_beta

    def quicklook(self) -> None:
        warnings.warn("Work in progress")
        quicklook_xarray(self.data, signal_var=f"signal_{self.channel}")

    def show(self) -> None:
        plt.show()

    def __repr__(self) -> str:
        return f"{self.data.attrs['system']} lidar data with {self.data.time.shape[0]} time measurements between {self.data.time[0]} and {self.data.time[-1]}"  # type: ignore



import numpy as np
import os



def offset_correction(physData, start, stop):
      arr = physData[start:stop] 
      return physData - np.mean(arr)

def smoothed_signal(data,filterWidth): 
      """ return a smoothed array based on the data input 
        Parameters
        ----------
        data : numpy array
        filterWidth : int
            The width of the filtering the larger the number the stronger the
            filtering
        """
      kernel_size = filterWidth
      kernel = np.ones(kernel_size) / kernel_size
      return np.convolve(data, kernel, mode='same')

